import logging
import time
import zmq
import orjson
import traceback
import threading
import queue
import os
import psutil
import signal
import concurrent.futures
import copy
import inspect
import functools
import sqlite3
import zlib
from contextlib import contextmanager

from . import NFP
from .client import NFPClient
from .keepalives import KeepAliver
from .security import generate_certificates
from .inventory import logging_config_producer
from typing import Any, Callable, Dict, List, Optional, Union
from norfab.models import NorFabEvent, Result
from norfab import models
from norfab.core.inventory import NorFabInventory
from jinja2.nodes import Include
from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, create_model

try:
    signal.signal(signal.SIGINT, signal.SIG_IGN)
except Exception:
    pass

log = logging.getLogger(__name__)

# --------------------------------------------------------------------------------------------
# NORFAB Worker Job Object
# --------------------------------------------------------------------------------------------


class Job:
    __slots__ = (
        "worker",
        "juuid",
        "client_address",
        "timeout",
        "args",
        "kwargs",
        "task",
        "client_input_queue",
    )

    def __init__(
        self,
        worker: object = None,
        juuid: str = None,
        client_address: str = None,
        timeout: int = None,
        args: list = None,
        kwargs: dict = None,
        task: str = None,
        client_input_queue: object = None,
    ):
        self.worker = worker
        self.juuid = juuid
        self.client_address = client_address
        self.timeout = timeout
        self.args = args or []
        self.kwargs = kwargs or {}
        self.task = task
        self.client_input_queue = client_input_queue

    def __str__(self):
        return self.juuid

    def event(self, message: str, **kwargs: Any):
        """
        Handles an event by forwarding it to the worker.

        Args:
            message (str): The message describing the event.
            **kwargs: Additional keyword arguments to include in the event.
        """
        kwargs.setdefault("task", self.task)
        if self.kwargs.get("progress", False) and self.juuid and self.worker:
            self.worker.event(
                message=message,
                juuid=self.juuid,
                client_address=self.client_address,
                **kwargs,
            )

    def stream(self, data: bytes) -> None:
        """
        Streams data to the broker.

        This method sends a message containing the client address, a unique
        identifier (UUID), a status code, and the provided data to the broker.

        Args:
            data (bytes): The data to be streamed to the broker.
        """
        msg = [
            self.client_address.encode("utf-8"),
            b"",
            self.juuid.encode("utf-8"),
            b"200",
            data,
        ]
        self.worker.send_to_broker(NFP.STREAM, msg)

    def wait_client_input(self, timeout: int = 10) -> Any:
        """
        Waits for input from the client within a specified timeout period if no item
        is available within the specified timeout, it returns `None`.

        Args:
            timeout (int, optional): The maximum time (in seconds) to wait for input

        Returns:
            Any: The item retrieved from the `client_input_queue`
        """
        try:
            return self.client_input_queue.get(block=True, timeout=timeout)
        except queue.Empty:
            pass

        return None


# --------------------------------------------------------------------------------------------
# NORFAB Worker Task Object
# --------------------------------------------------------------------------------------------

# Dictionary to store all tasks references
NORFAB_WORKER_TASKS = {}


class Task:
    """
    Validate is a class-based decorator that accept arguments, designed to validate the
    input arguments of a task function using a specified Pydantic model. It ensures that
    the arguments passed to the decorated function conform to the schema defined in the model.

    Attributes:
        model (BaseModel): A Pydantic model used to validate the function arguments.
        name (str): The name of the task, which is used to register the task for calling, by default
            set equal to the name of decorated function.
        result_model (BaseModel): A Pydantic model used to validate the function's return value.
        fastapi (dict): Dictionary with parameters for FastAPI `app.add_api_route` method
        mcp (dict): Dictionary with parameters for MCP `mcp.types.Tool` class

    Methods:
        __call__(function: Callable) -> Callable:
            Wraps the target function and validates its arguments before execution.

        merge_args_to_kwargs(args: List, kwargs: Dict) -> Dict:
            Merges positional arguments (`args`) and keyword arguments (`kwargs`) into a single
            dictionary, mapping positional arguments to their corresponding parameter names
            based on the function's signature.

        validate_input(args: List, kwargs: Dict) -> None:
            Validates merged arguments against Pydantic model. If validation fails,
            an exception is raised.

    Usage:
        @Task()(input=YourPydanticModel)
        def your_function(arg1, arg2, ...):
            # Function implementation
            pass

    Notes:
        - The decorator uses `inspect.getfullargspec` to analyze the function's signature
          and properly map arguments for validation.
    """

    def __init__(
        self,
        input: Optional[BaseModel] = None,
        output: Optional[BaseModel] = None,
        description: Optional[str] = None,
        fastapi: Optional[dict] = None,
        mcp: Optional[dict] = None,
    ) -> None:
        self.input = input
        self.output = output or Result
        self.description = description
        if fastapi is False:
            self.fastapi = False
        else:
            self.fastapi = fastapi or {}
        if mcp is False:
            self.mcp = False
        else:
            self.mcp = mcp or {}

    def __call__(self, function: Callable) -> Callable:
        """
        Decorator to register a function as a worker task with input/output
        validation and optional argument filtering.

        This method wraps the provided function, validates its input arguments
        and output, and registers it as a task. It also removes 'job' and
        'progress' keyword arguments if the wrapped function does not accept them.

        Side Effects:

            - Sets self.function, self.description, and self.name based on the provided function.
            - Initializes input model if not already set.
            - Updates the global NORFAB_WORKER_TASKS with the task schema.
        """
        self.function = function
        self.description = self.description or function.__doc__
        self.name = function.__name__

        if self.input is None:
            self.make_input_model()

        @functools.wraps(self.function)
        def wrapper(*args, **kwargs):
            # remove `job` argument if function does not expect it
            if self.is_need_argument(function, "job") is False:
                _ = kwargs.pop("job", None)

            # remove `progress` argument if function does not expect it
            if self.is_need_argument(function, "progress") is False:
                _ = kwargs.pop("progress", None)

            # validate input arguments
            self.validate_input(args, kwargs)

            ret = self.function(*args, **kwargs)

            # validate result
            self.validate_output(ret)

            return ret

        NORFAB_WORKER_TASKS.update(self.make_task_schema(wrapper))

        log.debug(
            f"{function.__module__} PID {os.getpid()} registered task '{function.__name__}'"
        )

        return wrapper

    def make_input_model(self):
        """
        Dynamically creates a Pydantic input model for the worker's function by inspecting its signature.

        This method uses `inspect.getfullargspec` to extract the function's argument names, default values,
        keyword-only arguments, and type annotations. It then constructs a dictionary of field specifications,
        giving preference to type annotations where available, and excluding special parameters such as 'self',
        'return', 'job', and any *args or **kwargs. The resulting specification is used to create a Pydantic
        model, which is assigned to `self.input`.

        The generated model used for input validation.
        """
        (
            fun_args,  # list of the positional parameter names
            fun_varargs,  # name of the * parameter or None
            fun_varkw,  # name of the ** parameter or None
            fun_defaults,  # tuple of default argument values of the last n positional parameters
            fun_kwonlyargs,  # list of keyword-only parameter names
            fun_kwonlydefaults,  # dictionary mapping kwonlyargs parameter names to default values
            fun_annotations,  # dictionary mapping parameter names to annotations
        ) = inspect.getfullargspec(self.function)

        # form a dictionary keyed by args with their default values
        args_with_defaults = dict(
            zip(reversed(fun_args or []), reversed(fun_defaults or []))
        )

        # form a dictionary keyed by args that has no defaults with values set to
        # (Any, None) tuple if make_optional is True else set to (Any, ...)
        args_no_defaults = {
            k: (Any, ...) for k in fun_args if k not in args_with_defaults
        }

        # form dictionary keyed by args with annotations and tuple values
        args_with_hints = {
            k: (v, args_with_defaults.get(k, ...)) for k, v in fun_annotations.items()
        }

        # form merged kwargs giving preference to type hint annotations
        merged_kwargs = {**args_no_defaults, **args_with_defaults, **args_with_hints}

        # form final dictionary of fields
        fields_spec = {
            k: v
            for k, v in merged_kwargs.items()
            if k not in ["self", "return", "job", fun_varargs, fun_varkw]
        }

        log.debug(
            f"NorFab worker {self.name} task creating Pydantic input "
            f"model using fields spec: {fields_spec}"
        )

        # create Pydantic model
        self.input = create_model(self.name, **fields_spec)

    def make_task_schema(self, wrapper) -> dict:
        """
        Generates a task schema dictionary for the current worker.

        Args:
            wrapper (Callable): The function wrapper to be associated with the task.

        Returns:
            dict: A dictionary containing the task's metadata, including:
                - function: The provided wrapper function.
                - module: The module name where the original function is defined.
                - schema: A dictionary with the following keys:
                    - name (str): The name of the task.
                    - description (str): The description of the task.
                    - inputSchema (dict): The JSON schema for the input model.
                    - outputSchema (dict): The JSON schema for the output model.
                    - fastapi: FastAPI-specific metadata.
                    - mcp: Model Context protocol metadata
        """
        input_json_schema = self.input.model_json_schema()
        _ = input_json_schema.pop("title")
        output_json_schema = self.output.model_json_schema()

        return {
            self.name: {
                "function": wrapper,
                "module": self.function.__module__,
                "schema": {
                    "name": str(self.name),
                    "description": self.description,
                    "inputSchema": input_json_schema,
                    "outputSchema": output_json_schema,
                    "fastapi": self.fastapi,
                    "mcp": self.mcp,
                },
            }
        }

    def is_need_argument(self, function: callable, argument: str) -> bool:
        """
        Determines whether a given argument name is required by the function.
        """
        fun_args, *_ = inspect.getfullargspec(function)
        return argument in fun_args

    def merge_args_to_kwargs(self, args: List, kwargs: Dict) -> Dict:
        """
        Merges positional arguments (`args`) and keyword arguments (`kwargs`)
        into a single dictionary.

        This function uses the argument specification of the decorated function
        to ensure that all arguments are properly combined into a dictionary.
        This is particularly useful for scenarios where **kwargs need to be passed
        to another function or model (e.g., for validation purposes).

        Arguments:
            args (list): A list of positional arguments passed to the decorated function.
            kwargs (dict): A dictionary of keyword arguments passed to the decorated function.

        Return:
            dict: A dictionary containing the merged arguments, where positional arguments
                  are mapped to their corresponding parameter names.
        """
        merged_kwargs = {}

        (
            fun_args,  # list of the positional parameter names
            fun_varargs,  # name of the * parameter or None
            *_,  # ignore the rest
        ) = inspect.getfullargspec(self.function)

        # "def foo(a, b):" - combine "foo(1, 2)" args with "a, b" fun_args
        args_to_kwargs = dict(zip(fun_args, args))

        # "def foo(a, *b):" - combine "foo(1, 2, 3)" 2|3 args with "*b" fun_varargs
        if fun_varargs:
            args_to_kwargs[fun_varargs] = args[len(fun_args) :]

        merged_kwargs = {**kwargs, **args_to_kwargs}

        # remove reference to self if decorating class method
        _ = merged_kwargs.pop("self", None)

        return merged_kwargs

    def validate_input(self, args: List, kwargs: Dict) -> None:
        """Function to validate provided arguments against model"""
        merged_kwargs = self.merge_args_to_kwargs(args, kwargs)
        log.debug(f"{self.name} validating input arguments: {merged_kwargs}")
        # if below step succeeds, kwargs passed model validation
        _ = self.input(**merged_kwargs)
        log.debug(
            f"Validated input kwargs: {merged_kwargs} for function {self.function} using model {self.input}"
        )

    def validate_output(self, ret: Result) -> None:
        if isinstance(ret, Result) and self.output:
            _ = self.output(**ret.model_dump())
        log.debug(f"Validated {self.name} task result.")


# --------------------------------------------------------------------------------------------
# NORFAB Worker Job Database
# --------------------------------------------------------------------------------------------


class JobDatabase:
    """
    Thread-safe SQLite database manager for worker jobs.

    Handles all job persistence operations with proper thread safety through
    connection-level locking and WAL mode for concurrent reads.

    Attributes:
        db_path (str): Path to the SQLite database file.
        _local (threading.local): Thread-local storage for database connections.
        _lock (threading.Lock): Lock for write operations to ensure thread safety.
    """

    def __init__(self, db_path: str, jobs_compress: bool = True):
        """
        Initialize the job database.

        Args:
            db_path (str): Path to the SQLite database file.
            jobs_compress (bool): If True, compress args, kwargs, and result_data fields. Defaults to True.
        """
        self.db_path = db_path
        self.jobs_compress = jobs_compress
        self._local = threading.local()
        self._lock = threading.Lock()
        self._initialize_database()

    def _get_connection(self) -> sqlite3.Connection:
        """
        Get a thread-local database connection.

        Returns:
            sqlite3.Connection: Thread-local database connection.
        """
        if not hasattr(self._local, "conn"):
            self._local.conn = sqlite3.connect(
                self.db_path, check_same_thread=False, timeout=30.0
            )
            self._local.conn.row_factory = sqlite3.Row
            # Enable WAL mode for better concurrent access
            self._local.conn.execute("PRAGMA journal_mode=WAL")
            self._local.conn.execute("PRAGMA synchronous=NORMAL")
            # performance PRagmas
            self._local.conn.execute("PRAGMA cache_size=5000")  # 5MB cache
            self._local.conn.execute("PRAGMA temp_store=MEMORY")  # Use RAM for temp
            self._local.conn.execute("PRAGMA mmap_size=30000000")  # Memory-mapped I/O
            self._local.conn.execute("PRAGMA query_only=False")  # Default
            self._local.conn.execute("PRAGMA busy_timeout=5000")  # 5s timeout

            # Verify JSON1 extension is available
            try:
                self._local.conn.execute("SELECT json('{}')").fetchone()
                log.debug("SQLite JSON1 extension is available")
            except sqlite3.OperationalError:
                log.warning(
                    "SQLite JSON1 extension not available - JSON queries will be limited"
                )
        return self._local.conn

    @contextmanager
    def _transaction(self, write: bool = False):
        """
        Context manager for database transactions with optional write locking.

        Args:
            write (bool): If True, acquire write lock for the transaction.

        Yields:
            sqlite3.Connection: Database connection.
        """
        conn = self._get_connection()
        if write:
            with self._lock:
                try:
                    yield conn
                    conn.commit()
                except Exception:
                    conn.rollback()
                    raise
        else:
            try:
                yield conn
            except Exception:
                raise

    def _compress_data(self, data: dict) -> bytes:
        """
        Serialize and optionally compress dictionary data to bytes for BLOB storage.

        Args:
            data (dict): Dictionary to serialize.

        Returns:
            bytes: zlib-compressed JSON bytes if compression enabled, otherwise raw JSON bytes.
        """
        if self.jobs_compress:
            return zlib.compress(orjson.dumps(data), level=6)
        else:
            return orjson.dumps(data)

    def _decompress_data(self, data_blob: bytes) -> dict:
        """
        Deserialize bytes retrieved from a BLOB column back to a dictionary.

        Args:
            data_blob (bytes): Raw or zlib-compressed JSON bytes.

        Returns:
            dict: Deserialized dictionary.
        """
        if self.jobs_compress:
            return orjson.loads(zlib.decompress(data_blob))
        else:
            return orjson.loads(data_blob)

    def _initialize_database(self):
        """Initialize the database schema."""
        with self._transaction(write=True) as conn:
            # Jobs table - args, kwargs, result_data stored as BLOBs
            conn.execute("""
                CREATE TABLE IF NOT EXISTS jobs (
                    uuid TEXT PRIMARY KEY,
                    client_address TEXT NOT NULL,
                    task TEXT NOT NULL,
                    args BLOB,
                    kwargs BLOB,
                    timeout INTEGER,
                    status TEXT DEFAULT 'PENDING',
                    received_timestamp TEXT NOT NULL,
                    started_timestamp TEXT,
                    completed_timestamp TEXT,
                    result_data BLOB,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """)

            # Events table
            conn.execute("""
                CREATE TABLE IF NOT EXISTS events (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    job_uuid TEXT NOT NULL,
                    message TEXT NOT NULL,
                    severity TEXT DEFAULT 'INFO',
                    task TEXT,
                    event_data TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    FOREIGN KEY (job_uuid) REFERENCES jobs(uuid) ON DELETE CASCADE
                )
            """)

            # Create indexes for performance
            conn.execute("CREATE INDEX IF NOT EXISTS idx_jobs_status ON jobs(status)")
            conn.execute("CREATE INDEX IF NOT EXISTS idx_jobs_task ON jobs(task)")
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_client ON jobs(client_address)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_received ON jobs(received_timestamp)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_jobs_created ON jobs(created_at)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_events_job_uuid ON events(job_uuid)"
            )
            conn.execute(
                "CREATE INDEX IF NOT EXISTS idx_events_severity ON events(severity)"
            )

    def add_job(
        self,
        uuid: str,
        client_address: str,
        task: str,
        args: list,
        kwargs: dict,
        timeout: int,
        timestamp: str,
    ) -> None:
        """
        Add a new job to the database.

        Args:
            uuid (str): Job UUID.
            client_address (str): Client address.
            task (str): Task name.
            args (list): Task arguments.
            kwargs (dict): Task keyword arguments.
            timeout (int): Job timeout.
            timestamp (str): Received timestamp.
        """
        with self._transaction(write=True) as conn:
            # Serialize args and kwargs to BLOB
            compressed_args = self._compress_data({"args": args})
            compressed_kwargs = self._compress_data({"kwargs": kwargs})

            conn.execute(
                """
                INSERT INTO jobs (uuid, client_address, task, args, kwargs, timeout,
                                 status, received_timestamp)
                VALUES (?, ?, ?, ?, ?, ?, 'PENDING', ?)
            """,
                (
                    uuid,
                    client_address,
                    task,
                    compressed_args,
                    compressed_kwargs,
                    timeout,
                    timestamp,
                ),
            )

    def get_next_pending_job(self) -> tuple:
        """
        Get the next pending job and mark it as STARTED.

        Returns:
            tuple: (uuid, received_timestamp) or None if no pending jobs.
        """
        with self._transaction(write=True) as conn:
            # order jobs by their creation timestamp in ascending order - oldest first
            cursor = conn.execute("""
                SELECT uuid, received_timestamp FROM jobs
                WHERE status = 'PENDING'
                ORDER BY created_at ASC
                LIMIT 1
            """)
            row = cursor.fetchone()
            if row:
                uuid = row["uuid"]
                conn.execute(
                    """
                    UPDATE jobs SET status = 'STARTED', started_timestamp = ?
                    WHERE uuid = ?
                """,
                    (time.ctime(), uuid),
                )
                return uuid, row["received_timestamp"]
            return None

    def complete_job(self, uuid: str, result_data: dict) -> None:
        """
        Mark a job as completed and store its result.

        Args:
            uuid (str): Job UUID.
            result_data (dict): Result data as dictionary.
        """
        with self._transaction(write=True) as conn:
            conn.execute(
                """
                UPDATE jobs
                SET status = 'COMPLETED', completed_timestamp = ?, result_data = ?
                WHERE uuid = ?
            """,
                (time.ctime(), self._compress_data(result_data), uuid),
            )

    def fail_job(self, uuid: str, result_data: dict) -> None:
        """
        Mark a job as failed and store its result.

        Args:
            uuid (str): Job UUID.
            result_data (dict): Result data as dictionary.
        """
        with self._transaction(write=True) as conn:
            conn.execute(
                """
                UPDATE jobs
                SET status = 'FAILED', completed_timestamp = ?, result_data = ?
                WHERE uuid = ?
            """,
                (time.ctime(), self._compress_data(result_data), uuid),
            )

    def get_job_info(
        self,
        uuid: str,
        include_result: bool = False,
        include_events: bool = False,
    ) -> dict:
        """
        Get comprehensive job information including status, execution data, and optionally result data and events.

        Args:
            uuid (str): Job UUID.
            include_result (bool): If True, include result_data in the response. Defaults to False.
            include_events (bool): If True, include job events. Defaults to False.

        Returns:
            dict: Job information with the following fields:
                - uuid: Job UUID
                - status: Job status (PENDING, STARTED, COMPLETED, FAILED, WAITING_CLIENT_INPUT)
                - received_timestamp: When job was received
                - started_timestamp: When job started execution
                - completed_timestamp: When job completed
                - client_address: Client address
                - task: Task name
                - args: Parsed task arguments list
                - kwargs: Parsed task keyword arguments dict
                - timeout: Job timeout

            If include_result=True, also includes:
                - result_data: Result data dictionary (if available)

            If include_events=True, also includes:
                - job_events: List of event dictionaries

            Returns None if job not found.
        """
        with self._transaction(write=False) as conn:
            # Build SELECT clause based on requested fields
            select_fields = [
                "uuid",
                "status",
                "received_timestamp",
                "started_timestamp",
                "completed_timestamp",
                "client_address",
                "task",
                "args",
                "kwargs",
                "timeout",
            ]

            if include_result:
                select_fields.append("result_data")

            query = f"""
                SELECT {', '.join(select_fields)}
                FROM jobs WHERE uuid = ?
            """

            cursor = conn.execute(query, (uuid,))
            row = cursor.fetchone()

            if not row:
                return None

            result = dict(row)
            result.setdefault("result_data", None)
            result.setdefault("job_events", [])

            # Deserialize args and kwargs from BLOB
            result["args"] = self._decompress_data(row["args"]).get("args", [])
            result["kwargs"] = self._decompress_data(row["kwargs"]).get("kwargs", {})

            # Include result data if requested
            if include_result and row["result_data"]:
                result["result_data"] = self._decompress_data(row["result_data"])

            # Include events if requested
            if include_events:
                result["job_events"] = self.get_job_events(uuid)

            return result

    def add_event(
        self, job_uuid: str, message: str, severity: str, task: str, event_data: dict
    ) -> None:
        """
        Add an event for a job.

        Args:
            job_uuid (str): Job UUID.
            message (str): Event message.
            severity (str): Event severity.
            task (str): Task name.
            event_data (dict): Event data dictionary.
        """
        with self._transaction(write=True) as conn:
            conn.execute(
                """
                INSERT INTO events (job_uuid, message, severity, task, event_data)
                VALUES (?, ?, ?, ?, ?)
            """,
                (
                    job_uuid,
                    message,
                    severity,
                    task,
                    orjson.dumps(event_data).decode("utf-8"),
                ),
            )

    def get_job_events(self, uuid: str) -> list:
        """
        Get all events for a job.

        Args:
            uuid (str): Job UUID.

        Returns:
            list: List of event dictionaries.
        """
        with self._transaction(write=False) as conn:
            cursor = conn.execute(
                """
                SELECT message, severity, task, event_data, created_at
                FROM events WHERE job_uuid = ?
                ORDER BY created_at ASC
            """,
                (uuid,),
            )
            return [
                {
                    "message": row["message"],
                    "severity": row["severity"],
                    "task": row["task"],
                    **orjson.loads(row["event_data"]),
                    "created_at": row["created_at"],
                }
                for row in cursor.fetchall()
            ]

    def list_jobs(
        self,
        pending: bool = True,
        completed: bool = True,
        task: str = None,
        last: int = None,
        client: str = None,
        uuid: str = None,
    ) -> list:
        """
        List jobs based on filters.

        Args:
            pending (bool): Include pending jobs.
            completed (bool): Include completed jobs.
            task (str): Filter by task name.
            last (int): Return only last N jobs.
            client (str): Filter by client address.
            uuid (str): Filter by specific UUID.

        Returns:
            list: List of job dictionaries.
        """
        with self._transaction(write=False) as conn:
            # Build query
            conditions = []
            params = []

            if uuid:
                conditions.append("uuid = ?")
                params.append(uuid)
            else:
                status_conditions = []
                if pending:
                    status_conditions.append("status IN ('PENDING', 'STARTED')")
                if completed:
                    status_conditions.append("status IN ('COMPLETED', 'FAILED')")
                if status_conditions:
                    conditions.append(f"({' OR '.join(status_conditions)})")

                if task:
                    conditions.append("task = ?")
                    params.append(task)
                if client:
                    conditions.append("client_address = ?")
                    params.append(client)

            where_clause = f"WHERE {' AND '.join(conditions)}" if conditions else ""
            query = f"""
                SELECT uuid, client_address, task, status,
                       received_timestamp, started_timestamp, completed_timestamp
                FROM jobs {where_clause}
                ORDER BY created_at DESC
            """

            if last:
                query += f" LIMIT {last}"

            cursor = conn.execute(query, params)
            return [dict(row) for row in cursor.fetchall()]

    def close(self):
        """Close all database connections."""
        if hasattr(self._local, "conn"):
            self._local.conn.close()
            delattr(self._local, "conn")


# --------------------------------------------------------------------------------------------
# NORFAB Worker watchdog Object
# --------------------------------------------------------------------------------------------


class WorkerWatchDog(threading.Thread):
    """
    Class to monitor worker performance.

    Attributes:
        worker (object): The worker instance being monitored.
        worker_process (psutil.Process): The process of the worker.
        watchdog_interval (int): Interval in seconds for the watchdog to check the worker's status.
        memory_threshold_mbyte (int): Memory usage threshold in megabytes.
        memory_threshold_action (str): Action to take when memory threshold is exceeded ("log" or "shutdown").
        runs (int): Counter for the number of times the watchdog has run.
        watchdog_tasks (list): List of additional tasks to run during each watchdog interval.

    Methods:
        check_ram(): Checks the worker's RAM usage and takes action if it exceeds the threshold.
        get_ram_usage(): Returns the worker's RAM usage in megabytes.
        run(): Main loop of the watchdog thread, periodically checks the worker's status and runs tasks.

    Args:
        worker (object): The worker object containing inventory attributes.
    """

    def __init__(self, worker):
        super().__init__()
        self.worker = worker
        self.worker_process = psutil.Process(os.getpid())

        # extract inventory attributes
        self.watchdog_interval = worker.inventory.get("watchdog_interval", 30)
        self.memory_threshold_mbyte = worker.inventory.get(
            "memory_threshold_mbyte", 1000
        )
        self.memory_threshold_action = worker.inventory.get(
            "memory_threshold_action", "log"
        )

        # initiate variables
        self.runs = 0
        self.watchdog_tasks = []

    def check_ram(self):
        """
        Checks the current RAM usage and performs an action if it exceeds the threshold.

        This method retrieves the current RAM usage and compares it to the predefined
        memory threshold. If the RAM usage exceeds the threshold, it performs an action
        based on the `memory_threshold_action` attribute. The possible actions are:

        - "log": Logs a warning message.
        - "shutdown": Raises a SystemExit exception to terminate the program.

        Raises:
            SystemExit: If the memory usage exceeds the threshold and the action is "shutdown".
        """
        mem_usage = self.get_ram_usage()
        if mem_usage > self.memory_threshold_mbyte:
            if self.memory_threshold_action == "log":
                log.warning(
                    f"{self.name} watchdog, '{self.memory_threshold_mbyte}' "
                    f"memory_threshold_mbyte exceeded, memory usage "
                    f"{mem_usage}MByte"
                )
            elif self.memory_threshold_action == "shutdown":
                raise SystemExit(
                    f"{self.name} watchdog, '{self.memory_threshold_mbyte}' "
                    f"memory_threshold_mbyte exceeded, memory usage "
                    f"{mem_usage}MByte, killing myself"
                )

    def get_ram_usage(self) -> float:
        """
        Get the RAM usage of the worker process.

        Returns:
            float: The RAM usage in megabytes.
        """
        return self.worker_process.memory_info().rss / 1024000

    def run(self):
        """
        Executes the worker's watchdog main loop, periodically running tasks and checking conditions.
        The method performs the following steps in a loop until the worker's exit event is set:

        1. Sleeps in increments of 0.1 seconds until the total sleep time reaches the watchdog interval.
        2. Runs built-in tasks such as checking RAM usage.
        3. Executes additional tasks provided by child classes.
        4. Updates the run counter.
        5. Resets the sleep counter to start the cycle again.

        Attributes:
            slept (float): The total time slept in the current cycle.
        """
        slept = 0
        while not self.worker.exit_event.is_set():
            # continue sleeping for watchdog_interval
            if slept < self.watchdog_interval:
                time.sleep(0.1)
                slept += 0.1
                continue

            # run built in tasks:
            self.check_ram()

            # run child classes tasks
            for task in self.watchdog_tasks:
                task()

            # update counters
            self.runs += 1

            slept = 0  # reset to go to sleep


# --------------------------------------------------------------------------------------------
# NORFAB worker
# --------------------------------------------------------------------------------------------


def _put(worker, put_queue, destroy_event):
    """
    Continuously processes items from the `put_queue` and updates the input queue
    of running jobs in the `worker` until the `destroy_event` is set.

    Args:
        worker (object): The worker instance that manages running jobs.
        put_queue (queue.Queue): A queue from which work items are retrieved.
        destroy_event (threading.Event): An event used to signal when the loop should stop

    Behavior:
        - The function retrieves work items from the `put_queue` with a timeout of 0.1 seconds.
        - If the queue is empty, it continues to the next iteration.
        - For each work item, it decodes the job data and updates the corresponding job's
          input queue in the `worker.running_jobs` dictionary.
    """
    while not destroy_event.is_set():
        try:
            work = put_queue.get(block=True, timeout=0.1)
        except queue.Empty:
            continue

        # Update job inputs
        suuid = None
        try:
            suuid = work[2].decode("utf-8")
            data = orjson.loads(work[3])
            worker.running_jobs[suuid].client_input_queue.put(data)
            log.debug(f"{worker.name} - '{suuid}' added job input")
        except Exception as e:
            log.error(
                f"{worker.name} - failed to update {suuid or '<unknown>'} job input: {e}",
                exc_info=True,
            )

        put_queue.task_done()


def _post(worker, post_queue, destroy_event):
    """
    Thread to receive POST requests and save them to database.

    Args:
        worker (Worker): The worker instance handling the request.
        post_queue (queue.Queue): The queue from which POST requests are received.
        destroy_event (threading.Event): Event to signal the thread to stop.

    Functionality:
        - Continuously processes POST requests from the queue until the destroy event is set.
        - Saves the request to the database.
        - Sends an acknowledgment back to the client.
    """
    while not destroy_event.is_set():
        try:
            work = post_queue.get(block=True, timeout=0.1)
        except queue.Empty:
            continue

        timestamp = time.ctime()
        client_address = work[0]
        suuid = work[2]
        data = orjson.loads(work[3])

        task = data.get("task")
        args = data.get("args", [])
        kwargs = data.get("kwargs", {})
        timeout = data.get("timeout", 60)

        # Add job to database
        try:
            worker.db.add_job(
                uuid=suuid.decode("utf-8"),
                client_address=client_address.decode("utf-8"),
                task=task,
                args=args,
                kwargs=kwargs,
                timeout=timeout,
                timestamp=timestamp,
            )
            log.debug(
                f"{worker.name} - '{suuid.decode('utf-8')}' job added to database"
            )
        except Exception as e:
            log.error(f"{worker.name} - failed to add job to database: {e}")
            post_queue.task_done()
            continue

        # ack job back to client
        worker.send_to_broker(
            NFP.RESPONSE,
            [
                client_address,
                b"",
                suuid,
                b"201",  # JOB CREATED
                orjson.dumps(
                    {
                        "worker": worker.name,
                        "uuid": suuid.decode("utf-8"),
                        "status": "ACCEPTED",
                        "service": worker.service.decode("utf-8"),
                    }
                ),
            ],
        )
        log.debug(
            f"{worker.name} - '{suuid.decode('utf-8')}' job, sent ACK back to client '{client_address.decode('utf-8')}'"
        )

        post_queue.task_done()


def _get(worker, get_queue, destroy_event):
    """
    Thread to receive GET requests and retrieve job status/results from the database.

    This function handles GET requests intelligently based on job status:
    - If job is PENDING or STARTED: Returns current status with timestamps
    - If job is COMPLETED or FAILED: Returns full job results
    - If job is not found: Returns 404 error

    Args:
        worker (Worker): The worker instance handling the request.
        get_queue (queue.Queue): The queue from which GET requests are received.
        destroy_event (threading.Event): Event to signal the thread to stop.
    """
    while not destroy_event.is_set():
        try:
            work = get_queue.get(block=True, timeout=0.1)
        except queue.Empty:
            continue

        client_address = work[0]
        suuid = work[2]
        uuid_str = suuid.decode("utf-8")
        reply = [client_address, b"", suuid]
        payload = {
            "worker": worker.name,
            "uuid": uuid_str,
            "service": worker.service.decode("utf-8"),
        }
        job_info = worker.db.get_job_info(uuid_str, include_result=True)

        if job_info is None:
            # Job not found
            status = b"404"
            payload["status"] = "JOB NOT FOUND"
        elif job_info["status"] in ("PENDING", "STARTED"):
            # Job is still in progress - return status with timestamps
            status = b"300"
            payload["status"] = job_info["status"]
        elif job_info["status"] == "WAITING_CLIENT_INPUT":
            # Job is still in progress - return status with timestamps
            status = b"102"
            payload["status"] = job_info["status"]
        elif job_info["status"] in ("COMPLETED", "FAILED"):
            result_dict = job_info.get("result_data")
            status = result_dict.get("status_code", "200").encode("utf-8")
            payload = result_dict["result"]
        else:
            # Unknown status
            status = b"500"
            payload["status"] = f"UNKNOWN STATUS: {job_info['status']}"

        reply.append(status)
        reply.append(orjson.dumps(payload))
        worker.send_to_broker(NFP.RESPONSE, reply)
        get_queue.task_done()


def _event(worker, event_queue, destroy_event):
    """
    Thread function to emit events to Clients.

    Args:
        worker (Worker): The worker instance that is emitting events.
        event_queue (queue.Queue): The queue from which events are retrieved.
        destroy_event (threading.Event): An event to signal the thread to stop.

    The function continuously retrieves events from the event_queue, processes them,
    and sends them to the broker until the destroy_event is set.
    """
    while not destroy_event.is_set():
        try:
            event_data = event_queue.get(block=True, timeout=0.1)
        except queue.Empty:
            continue
        uuid = event_data.pop("juuid")
        event = [
            event_data.pop("client_address").encode("utf-8"),
            b"",
            uuid.encode("utf-8"),
            b"200",
            orjson.dumps(
                {
                    "worker": worker.name,
                    "service": worker.service.decode("utf-8"),
                    "uuid": uuid,
                    **event_data,
                }
            ),
        ]
        worker.send_to_broker(NFP.EVENT, event)
        event_queue.task_done()


def recv(worker, destroy_event):
    """
    Thread to process receive messages from broker.

    This function runs in a loop, polling the worker's broker socket for messages every second.
    When a message is received, it processes the message based on the command type and places
    it into the appropriate queue or handles it accordingly. If the keepaliver thread is not
    alive, it logs a warning and attempts to reconnect to the broker.

    Args:
        worker (Worker): The worker instance that contains the broker socket and queues.
        destroy_event (threading.Event): An event to signal the thread to stop.

    Commands:
        - NFP.POST: Places the message into the post_queue.
        - NFP.DELETE: Places the message into the delete_queue.
        - NFP.GET: Places the message into the get_queue.
        - NFP.KEEPALIVE: Processes a keepalive heartbeat.
        - NFP.DISCONNECT: Attempts to reconnect to the broker.
        - Other: Logs an invalid input message.
    """
    while not destroy_event.is_set():
        # Poll socket for messages every 1000ms
        try:
            items = worker.poller.poll(1000)
        except KeyboardInterrupt:
            break  # Interrupted
        if items:
            with worker.socket_lock:
                msg = worker.broker_socket.recv_multipart()
            log.debug(f"{worker.name} - received '{msg}'")
            empty = msg.pop(0)  # noqa
            header = msg.pop(0)
            command = msg.pop(0)

            if command == NFP.POST:
                worker.post_queue.put(msg)
            elif command == NFP.DELETE:
                worker.delete_queue.put(msg)
            elif command == NFP.GET:
                worker.get_queue.put(msg)
            elif command == NFP.PUT:
                worker.put_queue.put(msg)
            elif command == NFP.KEEPALIVE:
                worker.keepaliver.received_heartbeat([header] + msg)
            elif command == NFP.DISCONNECT:
                worker.reconnect_to_broker()
            else:
                log.debug(
                    f"{worker.name} - invalid input, header '{header}', command '{command}', message '{msg}'"
                )

        if not worker.keepaliver.is_alive():
            log.warning(f"{worker.name} - '{worker.broker}' broker keepalive expired")
            worker.reconnect_to_broker()


class NFPWorker:
    """
    NFPWorker class is responsible for managing worker operations,
    including connecting to a broker, handling jobs,  and maintaining
    keepalive connections. It interacts with the broker using ZeroMQ
    and manages job queues and events.

    Args:
        inventory (NorFabInventory): The inventory object containing base directory information.
        broker (str): The broker address.
        service (str): The service name.
        name (str): The name of the worker.
        exit_event: The event used to signal the worker to exit.
        log_level (str, optional): The logging level. Defaults to None.
        log_queue (object, optional): The logging queue. Defaults to None.
        multiplier (int, optional): The multiplier value. Defaults to 6.
        keepalive (int, optional): The keepalive interval in milliseconds. Defaults to 2500.
    """

    keepaliver = None
    stats_reconnect_to_broker = 0

    def __init__(
        self,
        inventory: NorFabInventory,
        broker: str,
        service: str,
        name: str,
        exit_event: object,
        log_level: str = None,
        log_queue: object = None,
        multiplier: int = 6,
        keepalive: int = 2500,
    ):
        self.setup_logging(log_queue, log_level)
        self.inventory = inventory
        self.max_concurrent_jobs = max(1, inventory.get("max_concurrent_jobs", 5))
        self.jobs_compress = inventory.get("jobs_compress", True)
        self.broker = broker
        self.service = service.encode("utf-8") if isinstance(service, str) else service
        self.name = name
        self.exit_event = exit_event
        self.broker_socket = None
        self.multiplier = multiplier
        self.keepalive = keepalive
        self.socket_lock = (
            threading.Lock()
        )  # used for keepalives to protect socket object
        self.zmq_auth = self.inventory.broker.get("zmq_auth", True)
        self.build_message = NFP.MessageBuilder()

        # create base directories
        self.base_dir = os.path.join(
            self.inventory.base_dir, "__norfab__", "files", "worker", self.name
        )
        os.makedirs(self.base_dir, exist_ok=True)

        # Initialize SQLite database for job management
        db_path = os.path.join(self.base_dir, f"{self.name}.db")
        self.db = JobDatabase(db_path, jobs_compress=self.jobs_compress)

        # dictionary to store currently running jobs
        self.running_jobs = {}

        # create events and queues
        self.destroy_event = threading.Event()
        self.request_thread = None
        self.reply_thread = None
        self.recv_thread = None
        self.event_thread = None
        self.put_thread = None

        self.post_queue = queue.Queue(maxsize=0)
        self.get_queue = queue.Queue(maxsize=0)
        self.delete_queue = queue.Queue(maxsize=0)
        self.event_queue = queue.Queue(maxsize=0)
        self.put_queue = queue.Queue(maxsize=0)

        # generate certificates and create directories
        if self.zmq_auth is not False:
            generate_certificates(
                self.base_dir,
                cert_name=self.name,
                broker_keys_dir=os.path.join(
                    self.inventory.base_dir,
                    "__norfab__",
                    "files",
                    "broker",
                    "public_keys",
                ),
                inventory=self.inventory,
            )
            self.public_keys_dir = os.path.join(self.base_dir, "public_keys")
            self.secret_keys_dir = os.path.join(self.base_dir, "private_keys")

        self.ctx = zmq.Context()
        self.poller = zmq.Poller()
        self.reconnect_to_broker()

        self.client = NFPClient(
            self.inventory,
            self.broker,
            name=f"{self.name}-NFPClient",
            exit_event=self.exit_event,
        )

        self.tasks = NORFAB_WORKER_TASKS

    def setup_logging(self, log_queue, log_level: str) -> None:
        """
        Configures logging for the worker.

        This method sets up the logging configuration using a provided log queue and log level.
        It updates the logging configuration dictionary with the given log queue and log level,
        and then applies the configuration using `logging.config.dictConfig`.

        Args:
            log_queue (queue.Queue): The queue to be used for logging.
            log_level (str): The logging level to be set. If None, the default level is used.
        """
        logging_config_producer["handlers"]["queue"]["queue"] = log_queue
        if log_level is not None:
            logging_config_producer["root"]["level"] = log_level
        logging.config.dictConfig(logging_config_producer)

    def reconnect_to_broker(self):
        """
        Connect or reconnect to the broker.

        This method handles the connection or reconnection process to the broker.
        It performs the following steps:

        1. If there is an existing broker socket, it sends a disconnect message,
           unregisters the socket from the poller, and closes the socket.
        2. Creates a new DEALER socket and sets its identity.
        3. Loads the client's secret and public keys for CURVE authentication.
        4. Loads the server's public key for CURVE authentication.
        5. Connects the socket to the broker.
        6. Registers the socket with the poller for incoming messages.
        7. Sends a READY message to the broker to register the service.
        8. Starts or restarts the keepalive mechanism to maintain the connection.
        9. Increments the reconnect statistics counter.
        10. Logs the successful registration to the broker.
        """
        if self.broker_socket:
            self.send_to_broker(NFP.DISCONNECT)
            self.poller.unregister(self.broker_socket)
            self.broker_socket.close()

        self.broker_socket = self.ctx.socket(zmq.DEALER)
        self.broker_socket.setsockopt_unicode(zmq.IDENTITY, self.name, "utf8")
        self.broker_socket.linger = 0

        if self.zmq_auth is not False:
            # We need two certificates, one for the client and one for
            # the server. The client must know the server's public key
            # to make a CURVE connection.
            client_secret_file = os.path.join(
                self.secret_keys_dir, f"{self.name}.key_secret"
            )
            client_public, client_secret = zmq.auth.load_certificate(client_secret_file)
            self.broker_socket.curve_secretkey = client_secret
            self.broker_socket.curve_publickey = client_public

            # The client must know the server's public key to make a CURVE connection.
            server_public_file = os.path.join(self.public_keys_dir, "broker.key")
            server_public, _ = zmq.auth.load_certificate(server_public_file)
            self.broker_socket.curve_serverkey = server_public

        self.broker_socket.connect(self.broker)
        self.poller.register(self.broker_socket, zmq.POLLIN)

        # Register service with broker
        self.send_to_broker(NFP.READY)
        log.debug(f"{self.name} - NFP.READY sent to broker '{self.broker}'")

        # start keepalives
        if self.keepaliver is not None:
            self.keepaliver.restart(self.broker_socket)
        else:
            self.keepaliver = KeepAliver(
                address=None,
                socket=self.broker_socket,
                multiplier=self.multiplier,
                keepalive=self.keepalive,
                exit_event=self.destroy_event,
                service=self.service,
                whoami=NFP.WORKER,
                name=self.name,
                socket_lock=self.socket_lock,
            )
            self.keepaliver.start()

        self.stats_reconnect_to_broker += 1
        log.info(
            f"{self.name} - registered to broker at '{self.broker}', "
            f"service '{self.service.decode('utf-8')}'"
        )

    def send_to_broker(self, command, msg: list = None):
        """
        Send a message to the broker.

        Parameters:
            command (str): The command to send to the broker. Must be one of NFP.READY, NFP.DISCONNECT, NFP.RESPONSE, or NFP.EVENT.
            msg (list, optional): The message to send. If not provided, a default message will be created based on the command.

        Logs:
            Logs an error if the command is unsupported.
            Logs a debug message with the message being sent.

        Thread Safety:
            This method is thread-safe and uses a lock to ensure that the broker socket is accessed by only one thread at a time.
        """
        if command == NFP.READY:
            msg = self.build_message.worker_to_broker_ready(service=self.service)
        elif command == NFP.DISCONNECT:
            msg = self.build_message.worker_to_broker_disconnect(service=self.service)
        elif command == NFP.RESPONSE:
            msg = self.build_message.worker_to_broker_response(response_data=msg)
        elif command == NFP.EVENT:
            msg = self.build_message.worker_to_broker_event(event_data=msg)
        elif command == NFP.STREAM:
            msg = self.build_message.worker_to_broker_stream(data=msg)
        else:
            log.error(
                f"{self.name} - cannot send '{command}' to broker, command unsupported"
            )
            return

        log.debug(f"{self.name} - sending '{msg}'")

        with self.socket_lock:
            self.broker_socket.send_multipart(msg)

    def load_inventory(self) -> dict:
        """
        Load inventory data from the broker for this worker.

        This function retrieves inventory data from the broker service using the worker's name.
        It logs the received inventory data and returns the results if available.

        Returns:
            dict: The inventory data results if available, otherwise an empty dictionary.
        """
        inventory_data = self.client.mmi(
            "sid.service.broker", "get_inventory", kwargs={"name": self.name}
        )

        log.debug(f"{self.name} - worker received inventory data {inventory_data}")

        if inventory_data["results"]:
            return inventory_data["results"]
        else:
            return {}

    def worker_exit(self) -> None:
        """
        Method to override in child classes with a set of actions to perform on exit call.

        This method should be implemented by subclasses to define any cleanup or finalization
        tasks that need to be performed when the worker is exiting.
        """
        return None

    @Task(fastapi={"methods": ["GET"]})
    def get_inventory(self, job: Job) -> Result:
        """
        Retrieve the worker's inventory.

        This method should be overridden in child classes to provide the specific
        implementation for retrieving the inventory of a worker.

        Returns:
            Dict: A dictionary representing the worker's inventory.

        Raises:
            NotImplementedError: If the method is not overridden in a child class.
        """
        raise NotImplementedError

    @Task(fastapi={"methods": ["GET"]})
    def get_version(self) -> Result:
        """
        Retrieve the version report of the worker.

        This method should be overridden in child classes to provide the specific
        version report of the worker.

        Returns:
            Dict: A dictionary containing the version information of the worker.

        Raises:
            NotImplementedError: If the method is not overridden in a child class.
        """
        raise NotImplementedError

    def destroy(self, message=None):
        """
        Cleanly shuts down the worker by performing the following steps:

        1. Calls the worker_exit method to handle any worker-specific exit procedures.
        2. Sets the destroy_event to signal that the worker is being destroyed.
        3. Calls the destroy method on the client to clean up client resources.
        4. Joins all the threads (request_thread, reply_thread, event_thread, recv_thread) if they are not None, ensuring they have finished execution.
        5. Closes the database connections.
        6. Destroys the context with a linger period of 0 to immediately close all sockets.
        7. Stops the keepaliver to cease any keepalive signals.
        8. Logs an informational message indicating that the worker has been destroyed, including an optional message.

        Args:
            message (str, optional): An optional message to include in the log when the worker is destroyed.
        """
        self.worker_exit()
        self.destroy_event.set()
        self.client.destroy()

        # join all the threads
        if self.request_thread is not None:
            self.request_thread.join()
        if self.reply_thread is not None:
            self.reply_thread.join()
        if self.event_thread is not None:
            self.event_thread.join()
        if self.recv_thread:
            self.recv_thread.join()

        # close database
        self.db.close()

        self.ctx.destroy(0)

        # stop keepalives
        self.keepaliver.stop()

        log.info(f"{self.name} - worker destroyed, message: '{message}'")

    def is_url(self, url: str) -> bool:
        """
        Check if the given string is a URL supported by NorFab File Service.

        Args:
            url (str): The URL to check.

        Returns:
            bool: True if the URL supported by NorFab File Service, False otherwise.
        """
        return any(str(url).startswith(k) for k in ["nf://"])

    def fetch_file(
        self, url: str, raise_on_fail: bool = False, read: bool = True
    ) -> str:
        """
        Function to download file from broker File Sharing Service

        Args:
            url: file location string in ``nf://<filepath>`` format
            raise_on_fail: raise FIleNotFoundError if download fails
            read: if True returns file content, return OS path to saved file otherwise

        Returns:
            str: File content if read is True, otherwise OS path to the saved file.

        Raises:
            FileNotFoundError: If raise_on_fail is True and the download fails.
        """
        if not self.is_url(url):
            raise ValueError(f"Invalid URL format: {url}")

        result = self.client.fetch_file(url=url, read=read)
        status = result["status"]
        file_content = result["content"]
        msg = f"{self.name} - worker '{url}' fetch file failed with status '{status}'"

        if status == "200":
            return file_content
        elif raise_on_fail is True:
            raise FileNotFoundError(msg)
        else:
            log.error(msg)
            return None

    def jinja2_render_templates(
        self, templates: list[str], context: dict = None, filters: dict = None
    ) -> str:
        """
        Download (if needed) and render a list of Jinja2 templates with the given context and optional filters.

        Args:
            templates (list[str]): A list of Jinja2 template strings or NorFab file paths.
            context (dict): A dictionary containing the context variables for rendering the templates.
            filters (dict, optional): A dictionary of custom Jinja2 filters to be used during rendering.

        Returns:
            str: The rendered templates concatenated into a single string.
        """
        rendered = []
        filters = filters or {}
        context = context or {}
        for template in templates:
            j2env = Environment(loader="BaseLoader")
            j2env.filters.update(filters)  # add custom filters
            renderer = j2env.from_string(template)
            template = renderer.render(**context)
            # download template file and render it again
            if template.startswith("nf://"):
                filepath = self.jinja2_fetch_template(template)
                searchpath, filename = os.path.split(filepath)
                j2env = Environment(loader=FileSystemLoader(searchpath))
                j2env.filters.update(filters)  # add custom filters
                renderer = j2env.get_template(filename)
                rendered.append(renderer.render(**context))
            # template content is fully rendered
            else:
                rendered.append(template)

        return "\n".join(rendered)

    def jinja2_fetch_template(self, url: str) -> str:
        """
        Helper function to recursively download a Jinja2 template along with
        other templates referenced using "include" statements.

        Args:
            url (str): A URL in the format ``nf://file/path`` to download the file.

        Returns:
            str: The file path of the downloaded Jinja2 template.

        Raises:
            FileNotFoundError: If the file download fails.
            Exception: If Jinja2 template parsing fails.
        """
        filepath = self.fetch_file(url, read=False)
        if filepath is None:
            msg = f"{self.name} - file download failed '{url}'"
            raise FileNotFoundError(msg)

        # download Jinja2 template "include"-ed files
        content = self.fetch_file(url, read=True)
        j2env = Environment(loader="BaseLoader")
        try:
            parsed_content = j2env.parse(content)
        except Exception as e:
            msg = f"{self.name} - Jinja2 template parsing failed '{url}', error: '{e}'"
            raise Exception(msg)

        # run recursion on include statements
        for node in parsed_content.find_all(Include):
            include_file = node.template.value
            base_path = os.path.split(url)[0]
            self.jinja2_fetch_template(os.path.join(base_path, include_file))

        return filepath

    def event(
        self,
        message: str,
        juuid: str,
        task: str,
        client_address: str,
        **kwargs: Any,
    ) -> None:
        """
        Handles the creation and emission of an event.

        This method takes event data, processes it, and sends it to the event queue.
        It also saves the event data to the database for future reference.

        Args:
            message: The event message
            juuid: Job ID for which this event is generated
            task: Task name
            client_address: Client address
            **kwargs: Additional keyword arguments to be passed when creating a NorFabEvent instance

        Logs:
            Error: Logs an error message if the event data cannot be formed.
        """
        # construct NorFabEvent
        try:
            event_data = NorFabEvent(
                message=message,
                juuid=juuid,
                client_address=client_address,
                task=task,
                **kwargs,
            )
        except Exception as e:
            log.error(f"Failed to form event data, error {e}")
            return
        event_dict = event_data.model_dump(exclude_none=True)

        # emit event to the broker
        self.event_queue.put(event_dict)

        # check if need to emit log for this event
        if self.inventory["logging"].get("log_events", False):
            event_log = f"EVENT {self.name}:{task} - {message}"
            severity = event_dict.get("severity", "INFO")
            if severity == "INFO":
                log.info(event_log)
            elif severity == "DEBUG":
                log.debug(event_log)
            elif severity == "WARNING":
                log.warning(event_log)
            elif severity == "CRITICAL":
                log.critical(event_log)
            elif severity == "ERROR":
                log.error(event_log)

        # save event to database
        try:
            self.db.add_event(
                job_uuid=juuid,
                message=message,
                severity=event_dict.get("severity", "INFO"),
                task=task,
                event_data=event_dict,
            )
        except Exception as e:
            log.error(f"Failed to save event to database: {e}")

    @Task(fastapi={"methods": ["GET"]})
    def job_details(
        self,
        uuid: str = None,
        result: bool = True,
        events: bool = True,
    ) -> Result:
        """
        Method to get job details by UUID for completed jobs.

        Args:
            uuid (str): The job UUID to return details for.
            result (bool): If True, return job result.
            events (bool): If True, return job events.

        Returns:
            Result: A Result object with the job details.
        """
        job = self.db.get_job_info(
            uuid=uuid, include_result=result, include_events=events
        )

        if job:
            return Result(
                task=f"{self.name}:job_details",
                result=job,
            )
        else:
            raise RuntimeError(f"{self.name} - job with UUID '{uuid}' not found")

    @Task(fastapi={"methods": ["GET"]})
    def job_list(
        self,
        pending: bool = True,
        completed: bool = True,
        task: str = None,
        last: int = None,
        client: str = None,
        uuid: str = None,
    ) -> Result:
        """
        Method to list worker jobs completed and pending.

        Args:
            pending (bool): If True or None, return pending jobs. If False, skip pending jobs.
            completed (bool): If True or None, return completed jobs. If False, skip completed jobs.
            task (str, optional): If provided, return only jobs with this task name.
            last (int, optional): If provided, return only the last N completed and last N pending jobs.
            client (str, optional): If provided, return only jobs submitted by this client.
            uuid (str, optional): If provided, return only the job with this UUID.

        Returns:
            Result: Result object with a list of jobs.
        """
        jobs = self.db.list_jobs(
            pending=pending,
            completed=completed,
            task=task,
            last=last,
            client=client,
            uuid=uuid,
        )

        # Add worker and service information to each job
        for job in jobs:
            job["worker"] = self.name
            job["service"] = self.service.decode("utf-8")
            # Map database status to expected status names
            if job["status"] in ("PENDING", "STARTED"):
                job["status"] = "PENDING"
                job["completed_timestamp"] = None
            elif job["status"] in ("COMPLETED", "FAILED"):
                job["status"] = "COMPLETED"

        return Result(
            task=f"{self.name}:job_list",
            result=jobs,
        )

    @Task(
        fastapi={"methods": ["POST"]},
        input=models.WorkerEchoIn,
        output=models.WorkerEchoOut,
    )
    def echo(
        self,
        job: Job,
        raise_error: Union[bool, int, str] = None,
        sleep: int = None,
        *args: Any,
        **kwargs: Any,
    ) -> Result:
        """
        Echoes the job information and optional arguments, optionally sleeping or raising an error.

        Args:
            job (Job): The job instance containing job details.
            raise_error (str, optional): If provided, raises a RuntimeError with this message.
            sleep (int, optional): If provided, sleeps for the specified number of seconds.
            *args: Additional positional arguments to include in the result.
            **kwargs: Additional keyword arguments to include in the result.

        Returns:
            Result: An object containing job details and any provided arguments.

        Raises:
            RuntimeError: If `raise_error` is provided.
        """
        if sleep:
            time.sleep(sleep)
        if raise_error:
            raise RuntimeError(raise_error)
        return Result(
            result={
                "juuid": job.juuid,
                "client_address": job.client_address,
                "timeout": job.timeout,
                "task": job.task,
                "args": args,
                "kwargs": kwargs,
            }
        )

    @Task(fastapi={"methods": ["GET"]})
    def list_tasks(self, name: Union[None, str] = None, brief: bool = False) -> Result:
        """
        Lists tasks supported by worker.

        Args:
            name (str, optional): The name of a specific task to retrieve
            brief (bool, optional): If True, returns only the list of task names

        Returns:
            Results returned controlled by this logic:

                - If brief is True returns a list of task names
                - If name is provided returns list with single item - OpenAPI schema of the specified task
                - Otherwise returns a list of schemas for all tasks

        Raises:
            KeyError: If a specific task name is provided but not registered in NORFAB_WORKER_TASKS.
        """
        ret = Result()
        if brief:
            ret.result = list(sorted(NORFAB_WORKER_TASKS.keys()))
        elif name:
            if name not in NORFAB_WORKER_TASKS:
                raise KeyError(f"{name} - task not registered")
            ret.result = [NORFAB_WORKER_TASKS[name]["schema"]]
        else:
            ret.result = [t["schema"] for t in NORFAB_WORKER_TASKS.values()]
        return ret

    @Task(fastapi={"methods": ["GET"]})
    def delete_fetched_files(self, filepath) -> Result:
        return Result(result=self.client.delete_fetched_files(filepath))

    @Task(fastapi={"methods": ["GET"]})
    def get_worker_stats(self) -> Result:
        pass

    def start_threads(self) -> None:
        """
        Starts multiple daemon threads required for the worker's operation.

        This method initializes and starts the following threads:
            - request_thread: Handles posting requests using the _post function.
            - reply_thread: Handles receiving replies using the _get function.
            - event_thread: Handles event processing using the _event function.
            - recv_thread: Handles receiving data using the recv function.

        Each thread is started as a daemon and is provided with the necessary arguments,
        including queues and events as required.

        Returns:
            None
        """
        # Start threads
        self.request_thread = threading.Thread(
            target=_post,
            daemon=True,
            name=f"{self.name}_post_thread",
            args=(
                self,
                self.post_queue,
                self.destroy_event,
            ),
        )
        self.request_thread.start()
        self.reply_thread = threading.Thread(
            target=_get,
            daemon=True,
            name=f"{self.name}_get_thread",
            args=(self, self.get_queue, self.destroy_event),
        )
        self.reply_thread.start()
        self.event_thread = threading.Thread(
            target=_event,
            daemon=True,
            name=f"{self.name}_event_thread",
            args=(self, self.event_queue, self.destroy_event),
        )
        self.event_thread.start()
        self.put_thread = threading.Thread(
            target=_put,
            daemon=True,
            name=f"{self.name}_put_thread",
            args=(
                self,
                self.put_queue,
                self.destroy_event,
            ),
        )
        self.put_thread.start()
        # start receive thread after other threads
        self.recv_thread = threading.Thread(
            target=recv,
            daemon=True,
            name=f"{self.name}_recv_thread",
            args=(
                self,
                self.destroy_event,
            ),
        )
        self.recv_thread.start()

    def run_next_job(self, uuid: str):
        """
        Processes the next job from the database.

        This method performs the following steps:

        1. Loads job data from the database.
        2. Parses the job data to extract the task name, arguments, keyword arguments, and timeout.
        3. Executes the specified task method on the worker instance with the provided arguments.
        4. Handles any exceptions raised during task execution, logging errors and creating a failed Result object if needed.
        5. Saves the result of the job execution to the database.
        6. Marks the job as completed or failed in the database.

        Args:
            uuid (str): The job UUID to process.

        Raises:
            TypeError: If the executed task does not return a Result object.
        """
        log.debug(f"{self.name} - processing job request {uuid}")

        # Load job data from database
        job_data = self.db.get_job_info(uuid)
        if not job_data:
            log.error(f"{self.name} - job {uuid} not found in database")
            return

        client_address = job_data["client_address"]
        task = job_data["task"]
        args = job_data["args"]
        kwargs = job_data["kwargs"]
        timeout = job_data["timeout"]

        job = Job(
            worker=self,
            client_address=client_address,
            juuid=uuid,
            task=task,
            timeout=timeout,
            args=copy.deepcopy(args),
            kwargs=copy.deepcopy(kwargs),
            client_input_queue=queue.Queue(maxsize=0),
        )
        self.running_jobs[uuid] = job

        log.debug(
            f"{self.name} - doing task '{task}', timeout: '{timeout}', "
            f"args: '{args}', kwargs: '{kwargs}', client: '{client_address}', "
            f"job uuid: '{uuid}'"
        )

        # inform client that job started
        job.event(message="starting", status="running")

        # run the actual job
        try:
            task_started = time.ctime()
            result = NORFAB_WORKER_TASKS[task]["function"](
                self, *args, job=job, **kwargs
            )
            task_completed = time.ctime()
            if not isinstance(result, Result):
                raise TypeError(
                    f"{self.name} - task '{task}' did not return Result object, "
                    f"args: '{args}', kwargs: '{kwargs}', client: '{client_address}', "
                    f"job uuid: '{uuid}'; task returned '{type(result)}'"
                )
            result.task = result.task or f"{self.name}:{task}"
            result.status = result.status or "completed"
            result.juuid = result.juuid or uuid
            result.service = self.service.decode("utf-8")
            job_failed = False
        except Exception as e:
            task_completed = time.ctime()
            result = Result(
                task=f"{self.name}:{task}",
                errors=[traceback.format_exc()],
                messages=[f"Worker experienced error: '{e}'"],
                failed=True,
                juuid=uuid,
            )
            log.error(
                f"{self.name} - worker experienced error:\n{traceback.format_exc()}"
            )
            job_failed = True

        result.task_started = task_started
        result.task_completed = task_completed

        # Prepare result data for database storage as JSON-serializable dict
        result_data = {
            "client_address": client_address,
            "uuid": uuid,
            "status_code": "200",
            "result": {self.name: result.model_dump()},
            "worker": self.name,
            "service": self.service.decode("utf-8"),
        }

        # Save job result to database
        if job_failed:
            self.db.fail_job(uuid, result_data)
        else:
            self.db.complete_job(uuid, result_data)

        # remove job from running jobs
        _ = self.running_jobs.pop(uuid)

        # inform client that job completed
        job.event(message="completed", status="completed")

    def work(self):
        """
        Executes the main worker loop, managing job execution using a thread pool.

        This method starts necessary background threads, then enters a loop where it:

        - Queries the database for the next pending job.
        - Atomically marks the job as started in the database.
        - Submits the job to a thread pool executor for concurrent processing.
        - Waits briefly if no pending jobs are found.
        - Continues until either the exit or destroy event is set.

        Upon exit, performs cleanup by calling the `destroy` method with a status message.
        """

        self.start_threads()

        # start job threads and submit jobs in an infinite loop
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.max_concurrent_jobs,
            thread_name_prefix=f"{self.name}-job-thread",
        ) as executor:
            while not self.exit_event.is_set() and not self.destroy_event.is_set():
                # Get next pending job from database
                job_info = self.db.get_next_pending_job()

                if job_info is None:
                    # No pending jobs, wait a bit
                    time.sleep(0.1)
                    continue

                uuid, received_timestamp = job_info
                log.debug(f"{self.name} - submitting job {uuid} to executor")

                # Submit the job to workers
                executor.submit(self.run_next_job, uuid)

        # make sure to clean up
        self.destroy(
            f"{self.name} - exit event is set '{self.exit_event.is_set()}', "
            f"destroy event is set '{self.destroy_event.is_set()}'"
        )



from typing import Union, Any

from norfab.models import Result
from norfab.core.worker import Task, Job
from norfab.core.exceptions import UnsupportedPluginError
from nornir_salt.plugins.functions import FFun_functions, ResultSerializer
from nornir_napalm.plugins.tasks import napalm_get


class ParseTask:
    @Task(fastapi={"methods": ["POST"]})
    def parse(
        self,
        job: Job,
        plugin: str = "napalm",
        getters: Union[str, list] = "get_facts",
        template: str = None,
        commands: Union[str, list] = None,
        to_dict: bool = True,
        add_details: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Parse network device output using specified plugin and options.

        Args:
            job: NorFab Job object containing relevant metadata
            plugin (str): The plugin to use for parsing. Options are:

                - napalm - parse devices output using NAPALM getters
                - ttp - use TTP Templates to parse devices output
                - textfsm - use TextFSM templates to parse devices output

            getters (str): The getters to use with the "napalm" plugin.
            template (str): The template to use with the "ttp" or "textfsm" plugin.
            commands (list): The list of commands to run with the "ttp" or "textfsm" plugin.
            to_dict (bool): Whether to convert the result to a dictionary.
            add_details (bool): Whether to add details to the result.
            **kwargs: Additional keyword arguments to pass to the plugin.

        Returns:
            Result: A Result object containing the parsed data.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
        """
        filters = {k: kwargs.get(k) for k in list(kwargs.keys()) if k in FFun_functions}
        ret = Result(task=f"{self.name}:parse", result={} if to_dict else [])

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        if plugin == "napalm":
            nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors
            result = nr.run(task=napalm_get, getters=getters, **kwargs)
            ret.result = ResultSerializer(
                result, to_dict=to_dict, add_details=add_details
            )
            ret.failed = result.failed  # failed is true if any of the hosts failed
        elif plugin == "ttp":
            result = self.cli(
                job=job,
                commands=commands or [],
                run_ttp=template,
                **filters,
                **kwargs,
                to_dict=to_dict,
                add_details=add_details,
                plugin="netmiko",
            )
            ret.result = result.result
        elif plugin == "textfsm":
            result = self.cli(
                job=job,
                commands=commands,
                **filters,
                **kwargs,
                to_dict=to_dict,
                add_details=add_details,
                use_textfsm=True,
                textfsm_template=template,
                plugin="netmiko",
            )
            ret.result = result.result
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        return ret



import builtins

from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from .nornir_picle_shell_common import (
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    print_nornir_results,
)
from typing import Union, List
from nornir_salt.plugins.functions import TabulateFormatter
from picle.models import PipeFunctionsModel, Outputters


class NornirNetworkPing(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    use_host_name: StrictBool = Field(
        None,
        description="Ping host's name instead of host's hostname",
        json_schema_extra={"presence": True},
        alias="use-host-name",
    )
    count: StrictInt = Field(None, description="Number of pings to run")
    ping_timeout: StrictInt = Field(
        None,
        description="Time in seconds before considering each non-arrived reply permanently lost",
        alias="ping-timeout",
    )
    size: StrictInt = Field(None, description="Size of the entire packet to send")
    interval: Union[int, float] = Field(
        None, description="Interval to wait between pings"
    )
    payload: str = Field(None, description="Payload content if size is not set")
    sweep_start: StrictInt = Field(
        None,
        description="If size is not set, initial size in a sweep of sizes",
        alias="sweep-start",
    )
    sweep_end: StrictInt = Field(
        None,
        description="If size is not set, final size in a sweep of sizes",
        alias="sweep-end",
    )
    df: StrictBool = Field(
        None,
        description="Don't Fragment flag value for IP Header",
        json_schema_extra={"presence": True},
    )
    match: StrictBool = Field(
        None,
        description="Do payload matching between request and reply",
        json_schema_extra={"presence": True},
    )
    source: StrictStr = Field(None, description="Source IP address")

    class PicleConfig:
        outputter = print_nornir_results

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        kwargs["fun"] = "ping"
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if "ping_timeout" in kwargs:
            kwargs["timeout"] = kwargs.pop("ping_timeout")

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        result = NFCLIENT.run_job(
            "nornir",
            "network",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret


class NornirNetworkDns(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    use_host_name: StrictBool = Field(
        None,
        description="Ping host's name instead of host's hostname",
        json_schema_extra={"presence": True},
        alias="use-host-name",
    )
    servers: Union[StrictStr, List[StrictStr]] = Field(
        None, description="List of DNS servers to use"
    )
    dns_timeout: StrictInt = Field(
        None,
        description="Time in seconds before considering request lost",
        alias="dns-timeout",
    )
    ipv4: StrictBool = Field(
        None, description="Resolve 'A' record", json_schema_extra={"presence": True}
    )
    ipv6: StrictBool = Field(
        None, description="Resolve 'AAAA' record", json_schema_extra={"presence": True}
    )

    class PicleConfig:
        outputter = print_nornir_results

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        kwargs["fun"] = "resolve_dns"
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        if "dns_timeout" in kwargs:
            kwargs["timeout"] = kwargs.pop("dns_timeout")

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        result = NFCLIENT.run_job(
            "nornir",
            "network",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret


class NornirNetworkShell(BaseModel):
    ping: NornirNetworkPing = Field(None, description="Ping devices")
    dns: NornirNetworkDns = Field(None, description="Resolve DNS")

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-net]#"
        outputter = print_nornir_results
        pipe = PipeFunctionsModel



import logging
import os

from typing import Any
from norfab.models import Result
from norfab.core.worker import Task, Job
from norfab.core.exceptions import UnsupportedPluginError
from nornir_netmiko.tasks import netmiko_file_transfer
from nornir_salt.plugins.tasks import nr_test
from nornir_salt.plugins.functions import ResultSerializer

log = logging.getLogger(__name__)


class FileCopyTask:
    @Task(fastapi={"methods": ["POST"]})
    def file_copy(
        self,
        job: Job,
        source_file: str,
        plugin: str = "netmiko",
        to_dict: bool = True,
        add_details: bool = False,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> Result:
        """
        Task to transfer files to and from hosts using SCP.

        Args:
            job: NorFab Job object containing relevant metadata
            source_file (str): The path or URL of the source file to be copied in
                ``nf://path/to/file`` format
            plugin (str, optional): The plugin to use for file transfer. Supported plugins:

                - netmiko - uses `netmiko_file_transfer` task plugin.

            to_dict (bool, optional): Whether to return the result as a dictionary. Defaults to True.
            add_details (bool, optional): Whether to add detailed information to the result. Defaults to False.
            dry_run (bool, optional): If True, performs a dry run without making any changes. Defaults to False.
            **kwargs: Additional arguments to pass to the file transfer plugin.

        Returns:
            dict: The result of the file copy operation.

        Raises:
            UnsupportedPluginError: If the specified plugin is not supported.
        """
        timeout = job.timeout * 0.9
        ret = Result(task=f"{self.name}:file_copy", result={} if to_dict else [])

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # download file from broker
        if self.is_url(source_file):
            source_file_local = self.fetch_file(
                source_file, raise_on_fail=True, read=False
            )

        # decide on what send commands task plugin to use
        if plugin == "netmiko":
            task_plugin = netmiko_file_transfer
            kwargs["source_file"] = source_file_local
            kwargs.setdefault("socket_timeout", timeout / 5)
            kwargs.setdefault("dest_file", os.path.split(source_file_local)[-1])
        else:
            raise UnsupportedPluginError(f"Plugin '{plugin}' not supported")

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # run task
        log.debug(
            f"{self.name} - running file copy with arguments '{kwargs}', is dry run - '{dry_run}'"
        )
        if dry_run is True:
            result = nr.run(task=nr_test, name="file_copy_dry_run", **kwargs)
        else:
            with self.connections_lock:
                result = nr.run(task=task_plugin, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        self.watchdog.connections_update(nr, plugin)
        self.watchdog.connections_clean()

        return ret



import json
import builtins

from enum import Enum
from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from .nornir_picle_shell_common import (
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    print_nornir_results,
)
from typing import Union, Optional, List
from nornir_salt.plugins.functions import TabulateFormatter
from picle.models import PipeFunctionsModel, Outputters


class NrCliPluginNetmiko(BaseModel):
    # nornir_netmiko.tasks.netmiko_send_command arguments
    enable: Optional[StrictBool] = Field(
        None,
        description="Attempt to enter enable-mode",
        json_schema_extra={"presence": True},
    )
    use_timing: Optional[StrictBool] = Field(
        None,
        description="switch to send command timing method",
        json_schema_extra={"presence": True},
        alias="use-timing",
    )
    # netmiko send_command methods arguments
    expect_string: Optional[StrictStr] = Field(
        None,
        description="Regular expression pattern to use for determining end of output",
        alias="expect-string",
    )
    read_timeout: Optional[StrictInt] = Field(
        None,
        description="Maximum time to wait looking for pattern",
        alias="read-timeout",
    )
    auto_find_prompt: Optional[StrictBool] = Field(
        None,
        description="Use find_prompt() to override base prompt",
        alias="auto-find-prompt",
    )
    strip_prompt: Optional[StrictBool] = Field(
        None,
        description="Remove the trailing router prompt from the output",
        json_schema_extra={"presence": True},
        alias="strip-prompt",
    )
    strip_command: Optional[StrictBool] = Field(
        None,
        description="Remove the echo of the command from the output",
        json_schema_extra={"presence": True},
        alias="strip-command",
    )
    normalize: Optional[StrictBool] = Field(
        None,
        description="Ensure the proper enter is sent at end of command",
        json_schema_extra={"presence": True},
    )
    use_textfsm: Optional[StrictBool] = Field(
        None,
        description="Process command output through TextFSM template",
        json_schema_extra={"presence": True},
        alias="use-textfsm",
    )
    textfsm_template: Optional[StrictStr] = Field(
        None,
        description="Name of template to parse output with",
        alias="textfsm-template",
    )
    use_ttp: Optional[StrictBool] = Field(
        None,
        description="Process command output through TTP template",
        json_schema_extra={"presence": True},
        alias="use-ttp",
    )
    ttp_template: Optional[StrictBool] = Field(
        None, description="Name of template to parse output with", alias="ttp-template"
    )
    use_genie: Optional[StrictBool] = Field(
        None,
        description="Process command output through PyATS/Genie parser",
        json_schema_extra={"presence": True},
        alias="use-genie",
    )
    cmd_verify: Optional[StrictBool] = Field(
        None,
        description="Verify command echo before proceeding",
        json_schema_extra={"presence": True},
        alias="cmd-verify",
    )
    # nornir_salt.plugins.tasks.netmiko_send_commands arguments
    interval: Optional[StrictInt] = Field(
        None,
        description="Interval between sending commands",
    )
    use_ps: Optional[StrictBool] = Field(
        None,
        description="Use send command promptless method",
        json_schema_extra={"presence": True},
        alias="use-ps",
    )
    use_ps_timeout: Optional[StrictInt] = Field(
        None,
        description="Promptless mode absolute timeout",
        json_schema_extra={"presence": True},
        alias="use-ps-timeout",
    )
    split_lines: Optional[StrictBool] = Field(
        None,
        description="Split multiline string to individual commands",
        json_schema_extra={"presence": True},
        alias="split-lines",
    )
    new_line_char: Optional[StrictStr] = Field(
        None,
        description="Character to replace with new line before sending to device, default is _br_",
        alias="new-line-char",
    )
    repeat: Optional[StrictInt] = Field(
        None,
        description="Number of times to repeat the commands",
    )
    stop_pattern: Optional[StrictStr] = Field(
        None,
        description="Stop commands repeat if output matches provided glob pattern",
        alias="stop-pattern",
    )
    repeat_interval: Optional[StrictInt] = Field(
        None,
        description="Time in seconds to wait between repeating commands",
        alias="repeat-interval",
    )
    return_last: Optional[StrictInt] = Field(
        None,
        description="Returns requested last number of commands outputs",
        alias="return-last",
    )

    @staticmethod
    def run(*args, **kwargs):
        kwargs["plugin"] = "netmiko"
        return NornirCliShell.run(*args, **kwargs)

    class PicleConfig:
        outputter = print_nornir_results


class NrCliPluginScrapli(BaseModel):
    # nornir_scrapli.tasks.send_command arguments
    strip_prompt: Optional[StrictBool] = Field(
        None,
        description="Strip prompt from returned output",
        json_schema_extra={"presence": True},
        alias="strip-prompt",
    )
    failed_when_contains: Optional[StrictStr] = Field(
        None,
        description="String or list of strings indicating failure if found in response",
        alias="failed-when-contains",
    )
    timeout_ops: Optional[StrictInt] = Field(
        None, description="Timeout ops value for this operation", alias="timeout-ops"
    )
    # nornir_salt.plugins.tasks.scrapli_send_commands arguments
    interval: Optional[StrictInt] = Field(
        None,
        description="Interval between sending commands",
    )
    split_lines: Optional[StrictBool] = Field(
        None,
        description="Split multiline string to individual commands",
        json_schema_extra={"presence": True},
        alias="split-lines",
    )
    new_line_char: Optional[StrictStr] = Field(
        None,
        description="Character to replace with new line before sending to device, default is _br_",
        alias="new-line-char",
    )
    repeat: Optional[StrictInt] = Field(
        None,
        description="Number of times to repeat the commands",
    )
    stop_pattern: Optional[StrictStr] = Field(
        None,
        description="Stop commands repeat if output matches provided glob pattern",
        alias="stop-pattern",
    )
    repeat_interval: Optional[StrictInt] = Field(
        None,
        description="Time in seconds to wait between repeating commands",
        alias="repeat-interval",
    )
    return_last: Optional[StrictInt] = Field(
        None,
        description="Returns requested last number of commands outputs",
        alias="return-last",
    )

    @staticmethod
    def run(*args, **kwargs):
        kwargs["plugin"] = "scrapli"
        return NornirCliShell.run(*args, **kwargs)

    class PicleConfig:
        outputter = print_nornir_results


class NrCliPluginNapalm(BaseModel):
    # nornir_salt.plugins.tasks.napalm_send_commands arguments
    interval: Optional[StrictInt] = Field(
        None,
        description="Interval between sending commands",
    )
    split_lines: Optional[StrictBool] = Field(
        None,
        description="Split multiline string to individual commands",
        json_schema_extra={"presence": True},
        alias="split-lines",
    )
    new_line_char: Optional[StrictStr] = Field(
        None,
        description="Character to replace with new line before sending to device, default is _br_",
        alias="new-line-char",
    )

    @staticmethod
    def run(*args, **kwargs):
        kwargs["plugin"] = "napalm"
        return NornirCliShell.run(*args, **kwargs)

    class PicleConfig:
        outputter = print_nornir_results


class NrCliPlugins(BaseModel):
    netmiko: NrCliPluginNetmiko = Field(
        None, description="Use Netmiko plugin to configure devices"
    )
    scrapli: NrCliPluginScrapli = Field(
        None, description="Use Scrapli plugin to configure devices"
    )
    napalm: NrCliPluginNapalm = Field(
        None, description="Use NAPALM plugin to configure devices"
    )


class NornirCliShell(
    NorniHostsFilters, TabulateTableModel, NornirCommonArgs, ClientRunJobArgs
):
    commands: Union[StrictStr, List[StrictStr]] = Field(
        ...,
        description="List of commands to collect form devices",
        json_schema_extra={"multiline": True},
    )
    plugin: NrCliPlugins = Field(None, description="Connection plugin parameters")
    dry_run: Optional[StrictBool] = Field(
        None,
        description="Dry run the commands",
        json_schema_extra={"presence": True},
        alias="dry-run",
    )
    enable: Optional[StrictBool] = Field(
        None, description="Enter exec mode", json_schema_extra={"presence": True}
    )
    run_ttp: Optional[StrictStr] = Field(
        None, description="TTP Template to run", alias="run-ttp"
    )
    job_data: Optional[StrictStr] = Field(
        None, description="Path to YAML file with job data", alias="job-data"
    )

    @staticmethod
    def source_commands():
        completions = ClientRunJobArgs.walk_norfab_files()
        completions.append("load-terminal")
        return completions

    @staticmethod
    def source_run_ttp():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    def source_job_data():
        return ClientRunJobArgs.walk_norfab_files()

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result")
        nowait = kwargs.pop("nowait", False)

        # convert use_ps_timeout to timeout as use_ps expects "timeout" argument
        if kwargs.get("use_ps") and "use_ps_timeout" in kwargs:
            kwargs["timeout"] = kwargs.pop("use_ps_timeout")

        # extract job_data
        if kwargs.get("job_data") and not kwargs["job_data"].startswith("nf://"):
            kwargs["job_data"] = json.loads(kwargs["job_data"])

        # extract Tabulate arguments
        table = kwargs.pop("table", {})  # tabulate
        headers = kwargs.pop("headers", "keys")  # tabulate
        headers_exclude = kwargs.pop("headers_exclude", [])  # tabulate
        sortby = kwargs.pop("sortby", "host")  # tabulate
        reverse = kwargs.pop("reverse", False)  # tabulate

        if table:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False

        # run the job
        result = NFCLIENT.run_job(
            "nornir",
            "cli",
            workers=workers,
            args=args,
            kwargs=kwargs,
            uuid=uuid,
            timeout=timeout,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        result = log_error_or_result(result, verbose_result=verbose_result)

        # form table results
        if table:
            table_data = []
            for w_name, w_res in result.items():
                for item in w_res:
                    item["worker"] = w_name
                    table_data.append(item)
            ret = TabulateFormatter(
                table_data,
                tabulate=table,
                headers=headers,
                headers_exclude=headers_exclude,
                sortby=sortby,
                reverse=reverse,
            )
        else:
            ret = result

        return ret

    class PicleConfig:
        subshell = True
        prompt = "nf[nornir-cli]#"
        outputter = print_nornir_results
        pipe = PipeFunctionsModel


class CliPlugins(str, Enum):
    netmiko = "netmiko"
    scrapli = "scrapli"
    napalm = "napalm"


class NorniCliInput(
    NorniHostsFilters,
    TabulateTableModel,
    NornirCommonArgs,
    ClientRunJobArgs,
    NrCliPluginNetmiko,
    NrCliPluginScrapli,
    NrCliPluginNapalm,
    use_enum_values=True,
):
    plugin: CliPlugins = Field(None, description="Connection plugin parameters")
    workers: Union[StrictStr, List[StrictStr]] = Field(
        "all", description="Filter workers to target"
    )



import logging
import builtins

from picle.models import PipeFunctionsModel, Outputters
from pydantic import (
    BaseModel,
    StrictBool,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result
from .fastmcp_picle_shell_discover import Discover
from typing import Any

SERVICE = "fasmcp"
log = logging.getLogger(__name__)

# ---------------------------------------------------------------------------------------------
# FASTMCP SERVICE SHELL SHOW COMMANDS MODELS
# ---------------------------------------------------------------------------------------------


class FastMCPShowInventoryModel(ClientRunJobArgs):
    class PicleConfig:
        outputter = Outputters.outputter_yaml
        pipe = PipeFunctionsModel

    @staticmethod
    def run(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        result = NFCLIENT.run_job(
            "fastmcp",
            "get_inventory",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
            nowait=nowait,
        )
        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)


class FastMCPShowStatusModel(ClientRunJobArgs):
    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel

    @staticmethod
    def run(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        result = NFCLIENT.run_job(
            "fastmcp",
            "get_status",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
            nowait=nowait,
        )
        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)


class FastMCPShowToolsModel(ClientRunJobArgs):
    brief: StrictBool = Field(
        None,
        description="show tools names only",
        json_schema_extra={"presence": True},
    )
    service: StrictStr = Field(
        None,
        description="filter tools by service name",
    )
    name: StrictStr = Field(
        None,
        description="filter tools by name using glob pattern",
    )

    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel

    @staticmethod
    def run(*args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "any")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        result = NFCLIENT.run_job(
            "fastmcp",
            "get_tools",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
            nowait=nowait,
        )
        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)


class FastMCPShowCommandsModel(BaseModel):
    inventory: FastMCPShowInventoryModel = Field(
        None,
        description="show FastMCP inventory data",
    )
    version: Any = Field(
        None,
        description="show FastMCP service version report",
        json_schema_extra={
            "outputter": Outputters.outputter_yaml,
            "absolute_indent": 2,
            "function": "get_version",
        },
    )
    status: FastMCPShowStatusModel = Field(
        None,
        description="show FastMCP server status",
    )
    tools: FastMCPShowToolsModel = Field(
        None,
        description="show FastMCP server tools",
    )

    class PicleConfig:
        outputter = Outputters.outputter_nested
        pipe = PipeFunctionsModel

    @staticmethod
    def get_version(**kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        result = NFCLIENT.run_job("fastmcp", "get_version", workers=workers)
        return log_error_or_result(result)


# ---------------------------------------------------------------------------------------------
# FASTMCP SERVICE MAIN SHELL MODEL
# ---------------------------------------------------------------------------------------------


class FastMCPServiceCommands(BaseModel):
    discover: Discover = Field(
        None, description="Discover NorFab services tasks and create tools"
    )

    class PicleConfig:
        subshell = True
        prompt = "nf[fastmcp]#"



import logging

import copy

from typing import Union, Any
from norfab.core.worker import Task, Job
from norfab.models import Result
from .netbox_models import NetboxFastApiArgs
from .netbox_exceptions import UnsupportedNetboxVersion
from norfab.core.exceptions import UnsupportedServiceError

log = logging.getLogger(__name__)


class NetboxDevicesTasks:

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_devices(
        self,
        job: Job,
        filters: Union[None, list] = None,
        instance: Union[None, str] = None,
        dry_run: bool = False,
        devices: Union[None, list] = None,
        cache: Union[None, bool, str] = None,
    ) -> Result:
        """
        Retrieves device data from Netbox using the GraphQL API.

        Args:
            job: NorFab Job object containing relevant metadata
            filters (list, optional): A list of filter dictionaries to filter devices.
            instance (str, optional): The Netbox instance name.
            dry_run (bool, optional): If True, only returns the query content without executing it. Defaults to False.
            devices (list, optional): A list of device names to query data for.
            cache (Union[bool, str], optional): Cache usage options:

                - True: Use data stored in cache if it is up to date, refresh it otherwise.
                - False: Do not use cache and do not update cache.
                - "refresh": Ignore data in cache and replace it with data fetched from Netbox.
                - "force": Use data in cache without checking if it is up to date.

        Returns:
            dict: A dictionary keyed by device name with device data.

        Raises:
            Exception: If the GraphQL query fails or if there are errors in the query result.
        """
        instance = instance or self.default_instance
        ret = Result(task=f"{self.name}:get_devices", result={}, resources=[instance])
        cache = self.cache_use if cache is None else cache
        filters = filters or []
        devices = devices or []
        queries = {}  # devices queries
        device_fields = [
            "name",
            "last_updated",
            "custom_field_data",
            "tags {name}",
            "device_type {model}",
            "role {name}",
            "config_context",
            "tenant {name}",
            "platform {name}",
            "serial",
            "asset_tag",
            "site {name slug tags{name} }",
            "location {name}",
            "rack {name}",
            "status",
            "primary_ip4 {address}",
            "primary_ip6 {address}",
            "airflow",
            "position",
            "id",
        ]

        if cache == True or cache == "force":
            # retrieve last updated data from Netbox for devices
            last_updated_query = {
                f"devices_by_filter_{index}": {
                    "obj": "device_list",
                    "filters": filter_item,
                    "fields": ["name", "last_updated"],
                }
                for index, filter_item in enumerate(filters)
            }
            if devices:
                # use cache data without checking if it is up to date for cached devices
                if cache == "force":
                    for device_name in list(devices):
                        device_cache_key = f"get_devices::{device_name}"
                        if device_cache_key in self.cache:
                            devices.remove(device_name)
                            ret.result[device_name] = self.cache[device_cache_key]
                # query netbox last updated data for devices
                if self.nb_version[instance] >= (4, 4, 0):
                    dlist = '["{dl}"]'.format(dl='", "'.join(devices))
                    filters_dict = {"name": f"{{in_list: {dlist}}}"}
                else:
                    raise UnsupportedNetboxVersion(
                        f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                        f"minimum required version is {self.compatible_ge_v4}"
                    )
                last_updated_query["devices_by_devices_list"] = {
                    "obj": "device_list",
                    "filters": filters_dict,
                    "fields": ["name", "last_updated"],
                }
            last_updated = self.graphql(
                job=job,
                queries=last_updated_query,
                instance=instance,
                dry_run=dry_run,
            )
            last_updated.raise_for_status(f"{self.name} - get devices query failed")

            # return dry run result
            if dry_run:
                ret.result["get_devices_dry_run"] = last_updated.result
                return ret

            # try to retrieve device data from cache
            self.cache.expire()  # remove expired items from cache
            for devices_list in last_updated.result.values():
                for device in devices_list:
                    device_cache_key = f"get_devices::{device['name']}"
                    # check if cache is up to date and use it if so
                    if device_cache_key in self.cache and (
                        self.cache[device_cache_key].get("last_updated")
                        == device["last_updated"]
                        or cache == "force"
                    ):
                        ret.result[device["name"]] = self.cache[device_cache_key]
                        # remove device from list of devices to retrieve
                        if device["name"] in devices:
                            devices.remove(device["name"])
                    # cache old or no cache, fetch device data
                    elif device["name"] not in devices:
                        devices.append(device["name"])
        # ignore cache data, fetch data from netbox
        elif cache == False or cache == "refresh":
            queries = {
                f"devices_by_filter_{index}": {
                    "obj": "device_list",
                    "filters": filter_item,
                    "fields": device_fields,
                }
                for index, filter_item in enumerate(filters)
            }

        # fetch devices data from Netbox
        if devices or queries:
            if devices:
                if self.nb_version[instance] >= (4, 4, 0):
                    dlist = '["{dl}"]'.format(dl='", "'.join(devices))
                    filters_dict = {"name": f"{{in_list: {dlist}}}"}
                else:
                    raise UnsupportedNetboxVersion(
                        f"{self.name} - Netbox version {self.nb_version[instance]} is not supported, "
                        f"minimum required version is {self.compatible_ge_v4}"
                    )
                queries["devices_by_devices_list"] = {
                    "obj": "device_list",
                    "filters": filters_dict,
                    "fields": device_fields,
                }

            # send queries
            query_result = self.graphql(
                job=job, queries=queries, instance=instance, dry_run=dry_run
            )

            # check for errors
            if query_result.errors:
                msg = f"{self.name} - get devices query failed with errors:\n{query_result.errors}"
                raise Exception(msg)

            # return dry run result
            if dry_run:
                ret.result["get_devices_dry_run"] = query_result.result
                return ret

            # process devices data
            devices_data = query_result.result
            for devices_list in devices_data.values():
                for device in devices_list:
                    if device["name"] not in ret.result:
                        device_name = device.pop("name")
                        # cache device data
                        if cache != False:
                            cache_key = f"get_devices::{device_name}"
                            self.cache.set(cache_key, device, expire=self.cache_ttl)
                        # add device data to return result
                        ret.result[device_name] = device

        return ret

    @Task(
        fastapi={"methods": ["PATCH"], "schema": NetboxFastApiArgs.model_json_schema()}
    )
    def sync_device_facts(
        self,
        job: Job,
        instance: Union[None, str] = None,
        dry_run: bool = False,
        datasource: str = "nornir",
        timeout: int = 60,
        devices: Union[None, list] = None,
        batch_size: int = 10,
        branch: str = None,
        **kwargs: Any,
    ) -> Result:
        """
        Updates device facts in NetBox, this task updates this device attributes:

        - serial number

        Args:
            job: NorFab Job object containing relevant metadata
            instance (str, optional): The NetBox instance to use.
            dry_run (bool, optional): If True, no changes will be made to NetBox.
            datasource (str, optional): The data source to use. Supported datasources:

                - **nornir** - uses Nornir Service parse task to retrieve devices' data
                    using NAPALM `get_facts` getter

            timeout (int, optional): The timeout for the job execution. Defaults to 60.
            devices (list, optional): The list of devices to update.
            batch_size (int, optional): The number of devices to process in each batch.
            branch (str, optional): Branch name to use, need to have branching plugin installed,
                automatically creates branch if it does not exist in Netbox.
            **kwargs: Additional keyword arguments to pass to the datasource job.

        Returns:
            dict: A dictionary containing the results of the update operation.

        Raises:
            Exception: If a device does not exist in NetBox.
            UnsupportedServiceError: If the specified datasource is not supported.
        """
        devices = devices or []
        instance = instance or self.default_instance
        ret = Result(
            task=f"{self.name}:sync_device_facts",
            resources=[instance],
            dry_run=dry_run,
            diff={},
            result={},
        )
        nb = self._get_pynetbox(instance, branch=branch)
        kwargs["add_details"] = True

        if datasource == "nornir":
            # source hosts list from Nornir
            if kwargs:
                devices.extend(self.get_nornir_hosts(kwargs, timeout))
                devices = list(set(devices))
                job.event(f"Syncing {len(devices)} devices")
            # fetch devices data from Netbox
            nb_devices = self.get_devices(
                job=job,
                instance=instance,
                devices=copy.copy(devices),
                cache="refresh",
            ).result
            # remove devices that does not exist in Netbox
            for d in list(devices):
                if d not in nb_devices:
                    msg = f"'{d}' device does not exist in Netbox"
                    ret.errors.append(msg)
                    log.error(msg)
                    devices.remove(d)
            # iterate over devices in batches
            for i in range(0, len(devices), batch_size):
                kwargs["FL"] = devices[i : i + batch_size]
                kwargs["getters"] = "get_facts"
                job.event(f"retrieving facts for devices {', '.join(kwargs['FL'])}")
                data = self.client.run_job(
                    "nornir",
                    "parse",
                    kwargs=kwargs,
                    workers="all",
                    timeout=timeout,
                )

                # Collect devices to update in bulk
                devices_to_update = []

                for worker, results in data.items():
                    if results["failed"]:
                        msg = f"{worker} get_facts failed, errors: {'; '.join(results['errors'])}"
                        ret.errors.append(msg)
                        log.error(msg)
                        continue
                    for host, host_data in results["result"].items():
                        if host_data["napalm_get"]["failed"]:
                            msg = f"{host} facts update failed: '{host_data['napalm_get']['exception']}'"
                            ret.errors.append(msg)
                            log.error(msg)
                            continue

                        nb_device = nb_devices[host]

                        facts = host_data["napalm_get"]["result"]["get_facts"]
                        desired_state = {
                            "serial": facts["serial_number"],
                        }
                        current_state = {
                            "serial": nb_device["serial"],
                        }
                        # Compare and get fields that need updating
                        updates, diff = self.compare_netbox_object_state(
                            desired_state=desired_state,
                            current_state=current_state,
                        )

                        # Only update if there are changes
                        if updates:
                            updates["id"] = int(nb_device["id"])
                            devices_to_update.append(updates)
                            ret.diff[host] = diff

                        ret.result[host] = {
                            (
                                "sync_device_facts_dry_run"
                                if dry_run
                                else "sync_device_facts"
                            ): (updates if updates else "Device facts in sync")
                        }
                        if branch is not None:
                            ret.result[host]["branch"] = branch

                # Perform bulk update
                if devices_to_update and not dry_run:
                    try:
                        nb.dcim.devices.update(devices_to_update)
                    except Exception as e:
                        ret.errors.append(f"Bulk update failed: {e}")
        else:
            raise UnsupportedServiceError(
                f"'{datasource}' datasource service not supported"
            )

        return ret



import time
import threading
import logging

from . import NFP

log = logging.getLogger(__name__)


class KeepAliver:
    """
    Helper class to run keepalives between Broker and Workers in a consistent way.

    Args:
        address (str): Optional address to send keepalives to.
        socket: ZeroMQ socket to use to send keepalives to.
        multiplier (int): Number of keepalives before considered dead.
        keepalive (int): Interval between keepalives in milliseconds.
        exit_event (threading.Event): Global exit event signaled by NFAPI if set, stop sending keepalives.
        service (str): Name of the service to include in keepalives.
        whoami (str): Identifier e.g. NFP.WORKER or NFP.BROKER to use as keepalives header.
        name (str): Descriptive name to include in logs.
        socket_lock: Lock to synchronize access to the socket.

    Attributes:
        address (str): Address to send keepalives to.
        socket: ZeroMQ socket to use to send keepalives to.
        exit_event (threading.Event): Global exit event.
        destroy_event (threading.Event): Event used by worker to stop keepalives.
        keepalive (int): Interval between keepalives in milliseconds.
        multiplier (int): Number of keepalives before considered dead.
        service (str): Name of the service to include in keepalives.
        whoami (str): Identifier to use as keepalives header.
        name (str): Descriptive name to include in logs.
        socket_lock: Lock to synchronize access to the socket.
        started_at (float): Timestamp when keepalives started.
        keepalives_received (int): Number of keepalives received.
        keepalives_send (int): Number of keepalives sent.
        holdtime (float): Expiry time unless heartbeat is received.
        keepalive_at (float): Time to send the next keepalive.
        keepalive_thread (threading.Thread): Thread to run keepalives.

    Methods:
        start(): Start keepalives thread.
        stop(): Stop keepalives thread.
        run(): Send heartbeats at keepalive interval.
        received_heartbeat(msg): Update holdtime when a heartbeat is received.
        restart(socket): Restart keepalives with a new socket.
        is_alive(): Check if the other party is seen before expiry.
        show_holdtime(): Show remaining holdtime.
        show_alive_for(): Show duration since keepalives started.
    """

    def __init__(
        self,
        address: str,
        socket,
        multiplier: int,  # e.g. 6 times
        keepalive: int,  # e.g. 5000 ms
        exit_event: threading.Event,
        service: str,
        whoami: str,  # NFP.BROKER or NFP.WORKER
        name: str,
        socket_lock,
    ):
        self.address = address
        self.socket = socket
        self.exit_event = exit_event or threading.Event()
        self.destroy_event = (
            threading.Event()
        )  # destroy event, used by worker to stop keepalives
        self.keepalive = keepalive
        self.multiplier = multiplier
        self.service = service
        self.whoami = whoami
        self.name = f"{name}-keepaliver"
        self.socket_lock = socket_lock
        self.build_message = NFP.MessageBuilder()

        self.started_at = 0
        self.keepalives_received = 0
        self.keepalives_send = 0
        self.holdtime = (
            time.time() + 0.001 * self.multiplier * self.keepalive
        )  # expires at this point, unless heartbeat
        self.keepalive_at = (
            time.time() + 0.001 * self.keepalive
        )  # when to send keepalive

        self.keepalive_thread = threading.Thread(
            target=self.run, name=f"{self.name}_keepalives_thread", daemon=True
        )

    def start(self):
        """
        Start the keepalives thread and record the start time.

        This method initiates the keepalive thread and sets the `started_at`
        attribute to the current time.

        Returns:
            bool: True if the keepalive thread was successfully started.
        """
        self.keepalive_thread.start()
        self.started_at = time.time()
        return True

    def stop(self):
        """
        Stops the keepalive thread by setting the destroy event and joining the thread.

        This method first checks if the destroy event is not already set. If it is not set,
        it sets the destroy event to signal the keepalive thread to stop. Then, it waits
        for the keepalive thread to finish execution by calling join() on it.

        Returns:
            bool: True if the keepalive thread was successfully stopped.
        """
        if not self.destroy_event.is_set():
            self.destroy_event.set()
        self.keepalive_thread.join()
        return True

    def run(self):
        """
        Continuously send heartbeat messages at the specified keepalive interval.

        This method runs in a loop until either the `exit_event` or `destroy_event` is set.
        It sends a heartbeat message if the current time exceeds the `keepalive_at` timestamp.
        The message format depends on whether the `address` is specified.

        The method also handles exceptions that may occur during the sending of the message
        and logs the error. After sending a heartbeat, it updates the `keepalive_at` timestamp
        and increments the `keepalives_send` counter. The loop sleeps for 0.1 seconds between
        iterations.

        Raises:
            Exception: If an error occurs while sending the heartbeat message.
        """
        while not self.exit_event.is_set() and not self.destroy_event.is_set():
            if time.time() > self.keepalive_at:  # time to send heartbeat
                if self.whoami == NFP.WORKER:
                    msg = self.build_message.worker_to_broker_keepalive(
                        service=self.service
                    )
                elif self.whoami == NFP.BROKER:
                    msg = self.build_message.broker_to_worker_keepalive(
                        address=self.address,
                        service=self.service,
                    )
                with self.socket_lock:
                    try:
                        self.socket.send_multipart(msg)
                    except Exception as e:
                        log.error(
                            f"{self.name} - failed to send keepalive, error '{e}'"
                        )
                self.keepalive_at = time.time() + 0.001 * self.keepalive
                self.keepalives_send += 1
                log.debug(f"{self.name} - send keepalive '{msg}'")
            time.sleep(0.1)

    def received_heartbeat(self, msg):
        """
        Handles the reception of a heartbeat message from another party.

        This method updates the holdtime and increments the count of received keepalives.

        Args:
            msg (str): The heartbeat message received.
        """
        log.debug(f"{self.name} - received keepalive '{msg}'")
        self.keepalives_received += 1
        self.holdtime = time.time() + 0.001 * self.multiplier * self.keepalive

    def restart(self, socket):
        """
        Restart keepalives with a new socket.

        This method reinitializes the keepalive mechanism with a new socket. It resets
        the counters for received and sent keepalives, sets the start time to the current
        time, and calculates the holdtime and the next keepalive time based on the
        provided keepalive interval and multiplier.

        Args:
            socket: The new socket to be used for keepalives.
        """
        self.socket = socket
        self.keepalives_received = 0
        self.keepalives_send = 0
        self.started_at = time.time()
        self.holdtime = (
            time.time() + 0.001 * self.multiplier * self.keepalive
        )  # expires at this point, unless heartbeat
        self.keepalive_at = (
            time.time() + 0.001 * self.keepalive
        )  # when to send keepalive

    def is_alive(self):
        """
        Check if the other party is still alive based on the hold time.

        Returns:
            bool: True if the other party has been seen before the hold time expires, False otherwise.
        """
        return self.holdtime > time.time()

    def show_holdtime(self):
        """
        Calculate and return the remaining hold time.

        This method subtracts the current time from the holdtime attribute
        and rounds the result to one decimal place.

        Returns:
            float: The remaining hold time in seconds, rounded to one decimal place.
        """
        return round(self.holdtime - time.time(), 1)

    def show_alive_for(self):
        """
        Calculate the duration for which the instance has been alive.

        Returns:
            int: The number of seconds since the instance was started.
        """
        return int(time.time() - self.started_at)



import logging

from norfab.core.worker import Task, Job
from norfab.models import Result
from .netbox_models import NetboxFastApiArgs

log = logging.getLogger(__name__)


class NetboxBranchTasks:

    @Task(
        fastapi={"methods": ["DELETE"], "schema": NetboxFastApiArgs.model_json_schema()}
    )
    def delete_branch(
        self,
        job: Job,
        branch: str = None,
        instance: str = None,
    ) -> Result:
        """
        Deletes a branch with the specified name from the NetBox instance.

        Args:
            job (Job): The job context for the operation.
            branch (str, optional): The name of the branch to delete.
            instance (str, optional): The NetBox instance name.

        Returns:
            Result: An object containing the outcome of the deletion operation,
                including whether the branch was found and deleted.
        """
        instance = instance or self.default_instance
        ret = Result(
            task=f"{self.name}:delete_branch",
            result=None,
            resources=[instance],
        )
        nb = self._get_pynetbox(instance)

        job.event(f"Deleting branch '{branch}', Netbo instance '{instance}'")

        nb_branch = nb.plugins.branching.branches.get(name=branch)

        if nb_branch:
            nb_branch.delete()
            ret.result = True
            job.event(f"'{branch}' deleted from '{instance}' Netbox instance")
        else:
            msg = f"'{branch}' branch does not exist in '{instance}' Netbox instance"
            ret.result = None
            ret.messages.append(msg)
            job.event(msg)

        return ret



import logging

from typing import Any
from norfab.models import Result
from norfab.core.worker import Task, Job
from nornir_salt.plugins.functions import ResultSerializer

log = logging.getLogger(__name__)


class TaskTask:
    @Task(fastapi={"methods": ["POST"]})
    def task(self, job: Job, plugin: str, **kwargs: Any) -> Result:
        """
        Execute a Nornir task plugin.

        This method dynamically imports and executes a specified Nornir task plugin,
        using the provided arguments and keyword arguments. The `plugin` attribute
        can refer to a file to fetch from a file service, which must contain a function
        named `task` that accepts a Nornir task object as the first positional argument.

        Example of a custom task function file:

        ```python
        # define connection name for RetryRunner to properly detect it
        CONNECTION_NAME = "netmiko"

        # create task function
        def task(nornir_task_object, **kwargs):
            pass
        ```

        Note:
            The `CONNECTION_NAME` must be defined within the custom task function file if
            RetryRunner is in use. Otherwise, the connection retry logic is skipped, and
            connections to all hosts are initiated simultaneously up to the number of `num_workers`.

        Args:
            job: NorFab Job object containing relevant metadata
            plugin (str): The path to the plugin function to import, or a NorFab
                URL to download a custom task or template URL that resolves to a file.
            **kwargs (Any): Additional arguments to pass to the specified task plugin.

        Notes:
            - `add_details` (bool): If True, adds task execution details to the results.
            - `to_dict` (bool): If True, returns results as a dictionary. Defaults to True.
            - Host filters: keys matching `FFun_functions` are treated as Nornir host filters.

        Returns:
            Result: An instance of the Result class containing the task execution results.

        Raises:
            FileNotFoundError: If the specified plugin file cannot be downloaded.
        """
        # extract attributes
        add_details = kwargs.pop("add_details", False)  # ResultSerializer
        to_dict = kwargs.pop("to_dict", True)  # ResultSerializer
        ret = Result(task=f"{self.name}:task", result={} if to_dict else [])

        filtered_nornir, no_match_result = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            return ret

        # download task from broker and load it
        if plugin.startswith("nf://"):
            function_text = self.fetch_file(plugin)
            if function_text is None:
                raise FileNotFoundError(
                    f"{self.name} - '{plugin}' task plugin download failed"
                )

            # load task function running exec
            globals_dict = {}
            exec(function_text, globals_dict, globals_dict)
            task_function = globals_dict["task"]
        # import task function
        elif "." in plugin:
            # below same as "from nornir.plugins.tasks import task_fun as task_function"
            module_name, func_name = plugin.rsplit(".", 1)
            module = __import__(module_name, fromlist=[func_name])  # nosec
            task_function = getattr(module, func_name)
        else:
            raise RuntimeError(
                f"{self.name} - '{plugin}' task should either be a path "
                f"to a file or a module import string"
            )

        nr = self._add_processors(filtered_nornir, kwargs, job)  # add processors

        # run task
        log.debug(f"{self.name} - running Nornir task '{plugin}', kwargs '{kwargs}'")
        with self.connections_lock:
            result = nr.run(task=task_function, **kwargs)

        ret.failed = result.failed  # failed is true if any of the hosts failed
        ret.result = ResultSerializer(result, to_dict=to_dict, add_details=add_details)

        self.watchdog.connections_clean()

        return ret



import logging
import builtins

from picle.models import Outputters
from pydantic import (
    StrictBool,
    StrictStr,
    Field,
)
from ..common import ClientRunJobArgs, log_error_or_result, listen_events
from typing import Optional

log = logging.getLogger(__name__)


class Discover(ClientRunJobArgs):
    service: StrictStr = Field(
        "all", description="Service name to discover tasks and generate API for"
    )
    progress: Optional[StrictBool] = Field(
        True,
        description="Display progress events",
        json_schema_extra={"presence": True},
    )

    @staticmethod
    @listen_events
    def run(uuid, *args, **kwargs):
        NFCLIENT = builtins.NFCLIENT
        workers = kwargs.pop("workers", "all")
        timeout = kwargs.pop("timeout", 600)
        verbose_result = kwargs.pop("verbose_result", False)
        nowait = kwargs.pop("nowait", False)

        result = NFCLIENT.run_job(
            "fastapi",
            "discover",
            kwargs=kwargs,
            workers=workers,
            timeout=timeout,
            uuid=uuid,
            nowait=nowait,
        )

        if nowait:
            return result, Outputters.outputter_nested

        return log_error_or_result(result, verbose_result=verbose_result)

    class PicleConfig:
        outputter = Outputters.outputter_nested



import copy
import hashlib
import json
import logging
import yaml

from typing import Union, Any
from norfab.models import Result
from norfab.core.worker import Task, Job
from nornir_salt.plugins.tasks import nr_test
from nornir_salt.plugins.functions import FFun_functions, ResultSerializer
from nornir_salt.utils.pydantic_models import modelTestsProcessorSuite

log = logging.getLogger(__name__)


class TestTask:
    @Task(fastapi={"methods": ["POST"]})
    def test(
        self,
        job: Job,
        suite: Union[list, str],
        subset: str = None,
        dry_run: bool = False,
        remove_tasks: bool = True,
        failed_only: bool = False,
        return_tests_suite: bool = False,
        job_data: Any = None,
        extensive: bool = False,
        groups: list = None,
        **kwargs: Any,
    ) -> Result:
        """
        Function to test networks using a suite of tests.

        Args:
            job: NorFab Job object containing relevant metadata
            suite (Union[list, str]): URL Path to YAML file with tests or a list of test definitions
                or template URL that resolves to a file path.
            subset (str, optional): List or string with comma-separated non-case-sensitive glob
                patterns to filter tests by name. Ignored if dry_run is True.
            dry_run (bool, optional): If True, returns produced per-host tests suite content only.
            remove_tasks (bool, optional): If False, results will include other tasks output.
            failed_only (bool, optional): If True, returns test results for failed tests only.
            return_tests_suite (bool, optional): If True, returns rendered per-host tests suite
                content in addition to test results using a dictionary with ``results`` and ``suite`` keys.
            job_data (str, optional): URL to YAML file with data or dictionary/list of data
                to pass on to Jinja2 rendering context.
            extensive (bool, optional): return extensive results, equivalent to using these arguments:

                - remove_tasks = False
                - return_tests_suite = True
                - add_details = True
                - to_dict = False

            groups (list, optional): list of test group names to run

            **kwargs: Any additional arguments to pass on to the Nornir service task.

        Returns:
            dict: A dictionary containing the test results. If `return_tests_suite` is True,
                the dictionary will contain both the test results and the rendered test suite.

        Note:
            Result `failed` attribute is set to True if any of the tests failed for any of the hosts.

        Raises:
            RuntimeError: If there is an error in rendering the Jinja2 templates or loading the YAML.
        """
        tests = {}  # dictionary to hold per-host test suites
        # set extensive details flags
        if extensive is True:
            kwargs["add_details"] = True
            kwargs["to_dict"] = False
            remove_tasks = False
            return_tests_suite = True
        add_details = kwargs.get("add_details", False)  # ResultSerializer
        to_dict = kwargs.get("to_dict", True)  # ResultSerializer
        filters = {k: kwargs.get(k) for k in list(kwargs.keys()) if k in FFun_functions}
        ret = Result(task=f"{self.name}:test", result={} if to_dict else [])
        suites = {}  # dictionary to hold combined test suites

        filtered_nornir, ret = self.filter_hosts_and_validate(kwargs, ret)
        if ret.status == "no_match":
            if return_tests_suite is True:
                ret.result = {"test_results": [], "suite": {}}
            return ret

        # download job data
        job_data = self.load_job_data(job_data)

        # generate per-host test suites
        for host_name, host in filtered_nornir.inventory.hosts.items():
            # render suite using Jinja2
            try:
                rendered_suite = self.jinja2_render_templates(
                    templates=[suite],
                    context={
                        "host": host,
                        "norfab": self.client,
                        "job_data": job_data,
                        "netbox": self.add_jinja2_netbox(),
                    },
                    filters=self.add_jinja2_filters(),
                )
            except Exception as e:
                msg = f"{self.name} - '{suite}' Jinja2 rendering failed: '{type(e).__name__}:{e}'"
                raise RuntimeError(msg) from e
            # load suit using YAML
            try:
                tests[host_name] = yaml.safe_load(rendered_suite) or []
            except Exception as e:
                msg = f"{self.name} - '{suite}' YAML load failed: '{type(e).__name__}:{e}'"
                raise RuntimeError(msg) from e

        # validate tests suite
        try:
            _ = modelTestsProcessorSuite(tests=tests)
        except Exception as e:
            msg = f"{self.name} - '{suite}' suite validation failed: '{type(e).__name__}:{e}'"
            raise RuntimeError(msg) from e

        # download pattern, schema and custom function files
        for host_name in tests.keys():
            for index, item in enumerate(tests[host_name]):
                for k in ["pattern", "schema", "function_file"]:
                    if self.is_url(item.get(k)):
                        item[k] = self.fetch_file(
                            item[k], raise_on_fail=True, read=True
                        )
                        if k == "function_file":
                            item["function_text"] = item.pop(k)
                tests[host_name][index] = item

        # save per-host tests suite content before mutating it
        if return_tests_suite is True:
            return_suite = copy.deepcopy(tests)

        log.debug(f"{self.name} - running test '{suite}', is dry run - '{dry_run}'")
        # run dry run task
        if dry_run is True:
            result = filtered_nornir.run(
                task=nr_test, name="tests_dry_run", ret_data_per_host=tests
            )
            ret.result = ResultSerializer(
                result, to_dict=to_dict, add_details=add_details
            )
        # combine per-host tests in suites based on task and arguments
        # Why - to run tests using any nornir service tasks with various arguments
        else:
            for host_name, host_tests in tests.items():
                for test in host_tests:
                    dhash = hashlib.md5()
                    test_args = test.pop("norfab", {})
                    nrtask = test_args.get("nrtask", "cli")
                    assert nrtask in [
                        "cli",
                        "network",
                        "cfg",
                        "task",
                    ], f"{self.name} - unsupported NorFab Nornir Service task '{nrtask}'"
                    test_json = json.dumps(test_args, sort_keys=True).encode()
                    dhash.update(test_json)
                    test_hash = dhash.hexdigest()
                    suites.setdefault(test_hash, {"params": test_args, "tests": {}})
                    suites[test_hash]["tests"].setdefault(host_name, [])
                    suites[test_hash]["tests"][host_name].append(test)
            log.debug(
                f"{self.name} - combined per-host tests suites based on NorFab Nornir Service task and arguments:\n{suites}"
            )
            # run test suites collecting output from devices
            for tests_suite in suites.values():
                nrtask = tests_suite["params"].pop("nrtask", "cli")
                function_kwargs = {
                    **tests_suite["params"],
                    **kwargs,
                    **filters,
                    "tests": tests_suite["tests"],
                    "remove_tasks": remove_tasks,
                    "failed_only": failed_only,
                    "subset": subset,
                    "groups": groups,
                }
                result = getattr(self, nrtask)(
                    job=job, **function_kwargs
                )  # returns Result object
                # save test results into overall results
                if to_dict == True:
                    for host_name, host_res in result.result.items():
                        ret.result.setdefault(host_name, {})
                        ret.result[host_name].update(host_res)
                        # set return result failed to true if any of the tests failed
                        for test_res in host_res.values():
                            if add_details:
                                if test_res["result"] != "PASS":
                                    ret.failed = True
                            elif test_res != "PASS":
                                ret.failed = True
                else:
                    ret.result.extend(result.result)
                    # set return result failed to true if any of the tests failed
                    if any(r["result"] != "PASS" for r in result.result):
                        ret.failed = True

        # check if need to return tests suite content
        if return_tests_suite is True:
            ret.result = {"test_results": ret.result, "suite": return_suite}

        return ret



import json
from mdutils import MdUtils


def markdown_results(data: dict, service: str, task: str, kwargs: dict = None):
    kwargs = kwargs or {}
    if service == "nornir" and task == "test":
        return nornir_test_markdown(data, kwargs)
    else:
        return generic_markdown(data, kwargs)


def generic_markdown(data: dict, kwargs: dict = None):
    """
    Convert generic task results to markdown format with per-worker sections.
    """
    kwargs = kwargs or {}
    results = data.get("result_data", {})

    # Markdown generator
    md = MdUtils(file_name="task_results")

    # Extract service and task from first worker result for header
    service = "Unknown"
    task = "Unknown"

    # Process each worker's results
    if not results:
        md.new_paragraph(f"No results available: {data}")
        return md.file_data_text

    # Add main header
    first_worker = next(iter(results.values()))
    service = first_worker.get("service", "Unknown").title()
    task = first_worker.get("task", "Unknown").split(":")[-1].replace("_", " ").title()
    md.new_header(level=1, title=f"{service} service {task} results")

    # Add overall summary section
    md.new_header(level=2, title="Overall Summary")
    overall_status = data.get("status", "Unknown")
    status_indicator = "✅ Good" if overall_status == "COMPLETED" else "⚠️ Check Status"
    errors = data["errors"]
    workers_requested = data["workers_requested"]
    workers_completed = ", ".join(data["workers_completed"])
    workers_dispatched = ", ".join(data["workers_dispatched"])
    workers_pending = ", ".join(data["workers_started"])
    summary_text = (
        f"- **Overall Status:** {overall_status} {status_indicator}\n"
        f"- **Errors:** {len(errors)} error(s)\n"
        f"- **Workers Count:** {len(data['workers_dispatched'])}\n"
        f"- **Workers Requested:** {workers_requested}\n"
        f"- **Workers Done:** {workers_completed}\n"
        f"- **Workers Dispatched:** {workers_dispatched}\n"
        f"- **Workers Pending:** {workers_pending}\n"
        f"- **Error Details:** {', '.join(errors)}\n"
    )
    md.new_paragraph(summary_text)

    for worker_name, worker_data in results.items():
        # Create worker section header
        md.new_header(level=2, title=f"Worker: {worker_name}")

        # Prepare summary bullet points
        failed = worker_data.get("failed", False)
        status = worker_data.get("status", "Unknown")
        task_name = worker_data.get("task", "N/A")
        errors = worker_data.get("errors", [])
        messages = worker_data.get("messages", [])
        task_started = worker_data.get("task_started", "N/A")
        task_completed = worker_data.get("task_completed", "N/A")
        dry_run = worker_data.get("dry_run", False)

        # Create summary section with bullet points
        md.new_header(level=3, title="Summary")
        summary_text = (
            f"- **Task:** {task_name}\n"
            f"- **Status:** {status}\n"
            f"- **Failed:** {'❌ Yes' if failed else '✅ No'}\n"
            f"- **Started:** {task_started}\n"
            f"- **Completed:** {task_completed}\n"
            f"- **Dry Run:** {'Yes' if dry_run else 'No'}\n"
            f"- **Errors:** {', '.join(errors)}\n"
            f"- **Messages:** {', '.join(messages)}\n"
        )
        md.new_paragraph(summary_text)

        # Add results section as collapsible code block
        md.new_header(level=3, title="Results")
        result_data = worker_data.get("result", {})
        if result_data:
            results_json = json.dumps(result_data, indent=2, default=str)
            md.new_line(
                f"<details>\n"
                f"<summary>Click to expand results</summary>\n\n"
                f"```json\n{results_json}\n```\n\n"
                f"</details>\n"
            )
        else:
            md.new_paragraph("No results data available.")

    # Add debug section
    md.new_header(level=2, title="Debug Information")
    md.new_paragraph(
        "This section contains detailed debugging information for troubleshooting and inspection."
    )

    debug_html = (
        f"<details>\n"
        f"<summary>Input Arguments (kwargs)</summary>\n\n"
        f"```json\n{json.dumps(kwargs, indent=2, default=str)}\n```\n\n"
        f"</details>\n\n"
    )
    md.new_line(debug_html)

    return md.file_data_text


def nornir_test_markdown(data: dict, kwargs: dict = None):
    """
    Convert Nornir test task results to markdown format.
    """
    results = data.get("result_data", {})

    # Unified hosts test results dictionary structure:
    # {
    #     "hostname": {
    #         "suite": {...},           # Test suite definitions
    #         "commands": [...],        # Command outputs
    #         "results": [...],         # Test results
    #     }
    # }
    hosts_tests_results = {}

    # Table data collection
    table_text = [
        "#",
        "Host",
        "Test Name",
        "Result",
        "Exception",
    ]  # Header row for summary table
    total_rows = 1  # Counter for total table rows (includes header)
    table_rows = []  # List to collect and sort table rows
    total_failed = 0  # number of tests failed

    # HTML output buffers
    tests_details_html = ""  # HTML for hierarchical tests details section
    devices_commands_output_html = ""  # HTML for device command outputs section
    debug_html = ""  # HTML for debug section with kwargs and raw data
    hosts_test_suites_html = ""  # HTML for test suite definitions per host

    # Markdown generator
    md = MdUtils(file_name="nornir_test_results")

    if not results:
        md.new_paragraph("No results available.")
        return md.file_data_text

    # Process each worker's results and populate hosts_tests_results
    for worker_name, worker_data in results.items():
        worker_result = worker_data["result"]
        # Handle detailed result format (extensive=True)
        test_suite = worker_result.get("suite", {})
        test_results_list = worker_result.get("test_results", [])
        if test_results_list:
            for test_result in test_results_list:
                host = test_result.get("host", "unknown")
                name = test_result.get("name", "N/A")
                result = test_result.get("result", "N/A")
                exception = test_result.get("exception", "")
                hosts_tests_results.setdefault(
                    host, {"suite": {}, "commands": [], "results": []}
                )

                # Determine if this is a test result or command output
                # Test results have 'test' field and `success` field
                is_test_result = "success" in test_result and "test" in test_result

                # Only add test results to table
                if is_test_result:
                    status_icon = "✅ PASS" if result == "PASS" else "❌ FAIL"
                    table_rows.append([host, name, status_icon, exception or ""])
                    total_rows += 1
                    hosts_tests_results[host]["results"].append(test_result)
                # Store command outputs (non-test results)
                else:
                    hosts_tests_results[host]["commands"].append(test_result)
        # Store suite info
        if test_suite:
            for host, suite_tests in test_suite.items():
                hosts_tests_results.setdefault(
                    host, {"suite": {}, "commands": [], "results": []}
                )
                hosts_tests_results[host]["suite"] = suite_tests
        # Handle brief result format (extensive=False)
        if not all(k in worker_result for k in ["suite", "test_results"]):
            for device, tests in worker_result.items():
                for test_name, status in tests.items():
                    status_icon = "✅ PASS" if status == "PASS" else "❌ FAIL"
                    table_rows.append([device, test_name, status_icon, ""])
                    total_rows += 1

    # Prepare table data
    if total_rows > 1:
        # Sort table rows by host name, then by test name
        table_rows.sort(key=lambda x: (x[0], x[1]))
        index = 1
        # Add failed tests to the top
        for row in table_rows:
            if "FAIL" in row[2]:
                row.insert(0, index)
                table_text.extend(row)
                index += 1
                total_failed += 1
        # Append successful tests
        for row in table_rows:
            if "PASS" in row[2]:
                row.insert(0, index)
                table_text.extend(row)
                index += 1

    # Prepare Tests HTMLs
    if hosts_tests_results:
        for host in sorted(hosts_tests_results.keys()):
            host_results = hosts_tests_results[host]["results"]
            host_commands = hosts_tests_results[host]["commands"]
            host_suite = hosts_tests_results[host]["suite"]

            # Create hierarchical structure: Host > Test Name > Details
            if host_results:
                passed = sum(1 for test in host_results if test.get("result") == "PASS")
                failed = sum(1 for test in host_results if test.get("result") == "FAIL")

                tests_details_html += (
                    f'<details style="margin-left:20px;">\n'
                    f"<summary>{host} ({len(host_results)} tests, {passed} pass, {failed} fail)</summary>\n\n"
                )

                for test in sorted(host_results, key=lambda x: x.get("name", "")):
                    name = test.get("name", "N/A")
                    result = test.get("result", "N/A")
                    status_icon = "✅ PASS" if result == "PASS" else "❌ FAIL"

                    # form comments section
                    comments = []
                    for comment_line in test.get("comments", "").splitlines():
                        comments.append(f"\n> {comment_line}")
                    comments = "".join(comments)

                    tests_details_html += (
                        f'<details style="margin-left:40px;">\n'
                        f"<summary>{name} {status_icon}</summary>\n\n"
                        f"- **Task:** {test.get('task', 'N/A')}\n"
                        f"- **Test:** {test.get('test', 'N/A')}\n"
                        f"- **Criteria:** {test.get('criteria', 'N/A')}\n"
                        f"- **Result:** {result}\n"
                        f"- **Exception:** {test.get('exception', 'N/A')}\n"
                        f"- **Failed:** {test.get('failed', 'N/A')}\n"
                        f"- **Groups:** {', '.join(test.get('groups', []))}\n"
                        f"- **Description:** {test.get('description', '')}\n"
                        f"- **Comments:** {comments}\n\n"
                        f"</details>\n\n"
                    )

                tests_details_html += "</details>\n\n"

            # Prepare Device Output HTML
            if host_commands:
                cmds_html = ""
                for cmd in host_commands:
                    cmd_name = cmd.get("name", "Command")
                    cmd_output = str(cmd.get("result", ""))
                    cmds_html += (
                        f'<details style="margin-left:40px;">\n'
                        f"<summary>{host} - '{cmd_name}'</summary>\n\n"
                        f"```\n{cmd_output}\n```\n\n"
                        f"</details>\n"
                    )

                devices_commands_output_html += (
                    f'<details style="margin-left:20px;">\n'
                    f"<summary>{host} ({len(host_commands)} commands collected)</summary>\n\n"
                    f"{cmds_html}"
                    f"</details>\n\n"
                )

            # Prepare Test Suites HTML if available
            if host_suite:
                if not hosts_test_suites_html:
                    hosts_test_suites_html += (
                        '\n<details style="margin-left:20px;">\n'
                        "<summary>Test suites definitions for each host</summary>\n\n"
                    )
                suite_json = json.dumps(host_suite, indent=2)
                hosts_test_suites_html += (
                    f'<details style="margin-left:40px;">\n'
                    f"<summary>{host} ({len(host_suite)} tests)</summary>\n\n"
                    f"```json\n{suite_json}\n```\n\n"
                    f"</details>\n"
                )

        # complete Test Suites HTML section
        if hosts_test_suites_html:
            hosts_test_suites_html += "\n</details>\n"

    # Prepare Debug section HTML
    kwargs["suite"] = kwargs["suite"][:100]  # limit suite content
    debug_html = (
        f'<details style="margin-left:20px;">\n'
        f"<summary>Input Arguments (kwargs)</summary>\n\n"
        f"```json\n{json.dumps(kwargs, indent=2, default=str)}\n```\n\n"
        f"</details>\n\n"
    )

    # ==================== MD OUTPUT SECTION ====================

    # Add main header
    md.new_header(level=1, title="Tests Execution Report")

    # Tests Summary section
    md.new_header(level=2, title="Summary")
    if total_rows > 1:
        md.new_paragraph(
            f"Tests total - {total_rows - 1}, "
            f"failed - {total_failed}, "
            f"success rate - {int((total_failed / (total_rows - 1)) * 100)}%\n\n"
        )
        md.new_table(columns=5, rows=total_rows, text=table_text, text_align="left")
    else:
        md.new_paragraph("❌ Failed to produce summary results.\n\n")

    # Tests Details section
    md.new_header(level=2, title="Tests Details")
    if tests_details_html:
        md.new_paragraph(
            "Hierarchical expandable sections organized by device, then test name, containing complete test result details."
        )
        md.new_line(tests_details_html)
    else:
        md.new_paragraph(
            "❌ No detailed results available. Set `extensive` to `True` in input kwargs arguments.\n\n"
        )

    # Device Output section
    md.new_header(level=2, title="Device Outputs")
    if devices_commands_output_html:
        md.new_paragraph(
            "Expandable sections containing outputs collected during test execution for each host."
        )
        md.new_line(devices_commands_output_html)
    else:
        md.new_paragraph(
            "❌ No hosts outputs available. Set `extensive` to `True` in input kwargs arguments.\n\n"
        )

    # Debug section
    md.new_header(level=2, title="Debug")
    md.new_paragraph(
        "This section contains detailed debugging information for troubleshooting and inspection. Includes input arguments and complete raw results data used to produce sections above."
    )
    # Test Suites section (if available)
    if hosts_test_suites_html:
        md.new_line(hosts_test_suites_html)
    else:
        md.new_paragraph(
            "❌ No hosts test suites available. Set `extensive` to `True` in input kwargs arguments.\n\n"
        )
    # Debug details section
    md.new_line(debug_html)

    return md.file_data_text



"""
#### Sample Fakenos Worker Inventory

``` yaml
service: fakenos
broker_endpoint: "tcp://127.0.0.1:5555"

topologies:
  test_topology:
    default:
      password: user
      username: user
      port: [5000, 6000]
      server:
        plugin: ParamikoSshServer
        configuration:
          address: 127.0.0.1
          ssh_key_file: ./ssh-keys/ssh_host_rsa_key
          timeout: 1
      shell:
        configuration: {}
        plugin: CMDShell
      nos:
        configuration: {}
        plugin: cisco_ios
    hosts:
      R1:
        password: fakenos
        port: 5001
        username: fakenos
        server:
          plugin: ParamikoSshServer
          configuration:
            address: 0.0.0.0
        shell:
          plugin: CMDShell
          configuration:
            intro: Custom SSH Shell
      R2: {}
      core-router:
        replicas: 2
        port: [5000, 6000]
```

"""

import logging
import time
import threading

from norfab.core.worker import NFPWorker
from fakenos import FakeNOS

SERVICE = "fakenos"

log = logging.getLogger(__name__)


def whatch_stop(network, exit_event):
    while True:
        if exit_event.is_set():
            network.stop()
            break
        else:
            time.sleep(0.01)


class FakenosWorker(NFPWorker):
    inventory = None

    def __init__(
        self,
        inventory,
        broker,
        worker_name,
        exit_event=None,
        log_level="WARNING",
    ):
        super().__init__(inventory, broker, SERVICE, worker_name, exit_event, log_level)

        # get inventory from broker
        self.fakenos_inventory = self.load_inventory()

        # start FakeNOS Network
        self.network = FakeNOS(
            inventory={
                "hosts": self.fakenos_inventory.get("hosts", {}),
                "default": self.fakenos_inventory.get("default", {}),
            }
        )
        self.network.start()
        whatch_stop_thread = threading.Thread(
            target=whatch_stop,
            args=(
                self.network,
                exit_event,
            ),
        )
        whatch_stop_thread.start()

        log.info(f"{self.name} - Initiated")

    # ----------------------------------------------------------------------
    # FakeNOS Service Functions that exposed for calling
    # ----------------------------------------------------------------------

    def stop(self):
        self.network.stop()
        return "Done"

    def start(self):
        self.network.start()
        return "Done"

    def restart(self):
        self.network.stop()
        self.network.start()
        return "Done"

    def get_inventory(self):
        return dict(self.fakenos_inventory)

    def get_hosts(self):
        return list(self.network.hosts)



import logging

from typing import Union, List
from norfab.core.worker import Task, Job
from norfab.models import Result
from .netbox_models import NetboxFastApiArgs

log = logging.getLogger(__name__)


class NetboxNornirInventoryTasks:

    @Task(fastapi={"methods": ["GET"], "schema": NetboxFastApiArgs.model_json_schema()})
    def get_nornir_inventory(
        self,
        job: Job,
        filters: Union[None, list] = None,
        devices: Union[None, list] = None,
        instance: Union[None, str] = None,
        interfaces: Union[dict, bool] = False,
        connections: Union[dict, bool] = False,
        circuits: Union[dict, bool] = False,
        nbdata: bool = True,
        bgp_peerings: Union[dict, bool] = False,
        primary_ip: str = "ip4",
        cache: Union[bool, str] = None,
    ) -> Result:
        """
        Retrieve and construct Nornir inventory from NetBox data.

        Args:
            job: NorFab Job object containing relevant metadata
            filters (list, optional): List of filters to apply when retrieving devices from NetBox.
            devices (list, optional): List of specific devices to retrieve from NetBox.
            instance (str, optional): NetBox instance to use.
            interfaces (Union[dict, bool], optional): If True, include interfaces data
                    in the inventory. If a dict, use it as arguments for the get_interfaces method.
            connections (Union[dict, bool], optional): If True, include connections data
                    in the inventory. If a dict, use it as arguments for the get_connections method.
            circuits (Union[dict, bool], optional): If True, include circuits data in the
                    inventory. If a dict, use it as arguments for the get_circuits method.
            nbdata (bool, optional): If True, include a copy of NetBox device's data in the host's data.
            primary_ip (str, optional): Specify whether to use 'ip4' or 'ip6' for the primary
                    IP address. Defaults to 'ip4'.
            cache (Union[bool, str], optional): Cache usage options:

                - True: Use data stored in cache if it is up to date, refresh it otherwise.
                - False: Do not use cache and do not update cache.
                - "refresh": Ignore data in cache and replace it with data fetched from Netbox.
                - "force": Use data in cache without checking if it is up to date.

        Returns:
            dict: Nornir inventory dictionary containing hosts and their respective data.
        """
        hosts = {}
        filters = filters or []
        devices = devices or []
        inventory = {"hosts": hosts}
        ret = Result(task=f"{self.name}:get_nornir_inventory", result=inventory)

        # check Netbox status
        netbox_status = self.get_netbox_status(job=job, instance=instance)
        if netbox_status.result[instance or self.default_instance]["status"] is False:
            return ret

        # retrieve devices data
        nb_devices = self.get_devices(
            job=job, filters=filters, devices=devices, instance=instance, cache=cache
        )

        # form Nornir hosts inventory
        for device_name, device in nb_devices.result.items():
            host = device["config_context"].pop("nornir", {})
            host.setdefault("data", {})
            name = host.pop("name", device_name)
            hosts[name] = host
            # add platform if not provided in device config context
            if not host.get("platform"):
                if device["platform"]:
                    host["platform"] = device["platform"]["name"]
                else:
                    log.warning(f"{self.name} - no platform found for '{name}' device")
            # add hostname if not provided in config context
            if not host.get("hostname"):
                if device["primary_ip4"] and primary_ip in ["ip4", "ipv4"]:
                    host["hostname"] = device["primary_ip4"]["address"].split("/")[0]
                elif device["primary_ip6"] and primary_ip in ["ip6", "ipv6"]:
                    host["hostname"] = device["primary_ip6"]["address"].split("/")[0]
                else:
                    host["hostname"] = name
            # add netbox data to host's data
            if nbdata is True:
                host["data"].update(device)

        # return if no hosts found for provided parameters
        if not hosts:
            log.warning(f"{self.name} - no viable hosts returned by Netbox")
            return ret

        # add interfaces data
        if interfaces:
            # decide on get_interfaces arguments
            kwargs = interfaces if isinstance(interfaces, dict) else {}
            kwargs.setdefault("cache", cache)
            # add 'interfaces' key to all hosts' data
            for host in hosts.values():
                host["data"].setdefault("interfaces", {})
            # query interfaces data from netbox
            nb_interfaces = self.get_interfaces(
                job=job, devices=list(hosts), instance=instance, **kwargs
            )
            # save interfaces data to hosts' inventory
            while nb_interfaces.result:
                device, device_interfaces = nb_interfaces.result.popitem()
                hosts[device]["data"]["interfaces"] = device_interfaces

        # add connections data
        if connections:
            # decide on get_interfaces arguments
            kwargs = connections if isinstance(connections, dict) else {}
            kwargs.setdefault("cache", cache)
            # add 'connections' key to all hosts' data
            for host in hosts.values():
                host["data"].setdefault("connections", {})
            # query connections data from netbox
            nb_connections = self.get_connections(
                job=job, devices=list(hosts), instance=instance, **kwargs
            )
            # save connections data to hosts' inventory
            while nb_connections.result:
                device, device_connections = nb_connections.result.popitem()
                hosts[device]["data"]["connections"] = device_connections

        # add circuits data
        if circuits:
            # decide on get_interfaces arguments
            kwargs = circuits if isinstance(circuits, dict) else {}
            kwargs.setdefault("cache", cache)
            # add 'circuits' key to all hosts' data
            for host in hosts.values():
                host["data"].setdefault("circuits", {})
            # query circuits data from netbox
            nb_circuits = self.get_circuits(
                job=job, devices=list(hosts), instance=instance, **kwargs
            )
            # save circuits data to hosts' inventory
            while nb_circuits.result:
                device, device_circuits = nb_circuits.result.popitem()
                hosts[device]["data"]["circuits"] = device_circuits

        # add bgp peerings data
        if bgp_peerings:
            # decide on get_interfaces arguments
            kwargs = bgp_peerings if isinstance(bgp_peerings, dict) else {}
            kwargs.setdefault("cache", cache)
            # add 'bgp_peerings' key to all hosts' data
            for host in hosts.values():
                host["data"].setdefault("bgp_peerings", {})
            # query bgp_peerings data from netbox
            nb_bgp_peerings = self.get_bgp_peerings(
                job=job, devices=list(hosts), instance=instance, **kwargs
            )
            # save circuits data to hosts' inventory
            while nb_bgp_peerings.result:
                device, device_bgp_peerings = nb_bgp_peerings.result.popitem()
                hosts[device]["data"]["bgp_peerings"] = device_bgp_peerings

        return ret
